from collections.abc import Iterable

import tensorflow as tf
from tensorflow.keras.layers import Concatenate, Dense, Multiply, PReLU, ReLU
from tensorflow.python.ops import gen_sparse_ops

from spektral.layers.convolutional.conv import Conv
from spektral.layers.convolutional.message_passing import MessagePassing


class XENetConv(MessagePassing):
    r"""
    A XENet convolutional layer from the paper

      > [XENet: Using a new graph convolution to accelerate the timeline for protein design on quantum computers](https://www.biorxiv.org/content/10.1101/2021.05.05.442729v1)<br>
      > Jack B. Maguire, Daniele Grattarola, Eugene Klyshko, Vikram Khipple Mulligan, Hans Melo

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    For a version of this layer that supports batch mode, you can use
    `spektral.layers.XENetDenseConv` as a drop-in replacement.

    This layer computes for each node \(i\):
    $$
        \s_{ij} = \text{PReLU} \left( (\x_{i} \| \x_{j} \| \e_{ij} \| \e_{ji}) \W^{(s)} + \b^{(s)} \right) \\
        \s^{(\text{out})}_{i} = \sum\limits_{j \in \mathcal{N}(i)} \s_{ij} \\
        \s^{(\text{in})}_{i} = \sum\limits_{j \in \mathcal{N}(i)} \s_{ji} \\
        \x_{i}' = \sigma\left( (\x_{i} \| \s^{(\text{out})}_{i} \| \s^{(\text{in})}_{i}) \W^{(n)} + \b^{(n)} \right) \\
        \e_{ij}' = \sigma\left( \s_{ij} \W^{(e)} + \b^{(e)} \right)
    $$

    **Input**

    - Node features of shape `([batch], n_nodes, n_node_features)`;
    - Binary adjacency matrices of shape `([batch], n_nodes, n_nodes)`;
    - Edge features of shape `(num_edges, n_edge_features)`;

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `node_channels`.
    - Edge features with the same shape of the input, but the last dimension
    changed to `edge_channels`.

    **Arguments**

    - `stack_channels`: integer or list of integers, number of channels for the hidden layers;
    - `node_channels`: integer, number of output channels for the nodes;
    - `edge_channels`: integer, number of output channels for the edges;
    - `attention`: whether to use attention when aggregating the stacks;
    - `node_activation`: activation function for nodes;
    - `edge_activation`: activation function for edges;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        stack_channels,
        node_channels,
        edge_channels,
        attention: bool = True,
        node_activation=None,
        edge_activation=None,
        aggregate: str = "sum",
        use_bias: bool = True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.stack_channels = stack_channels
        self.node_channels = node_channels
        self.edge_channels = edge_channels
        self.attention = attention
        self.node_activation = node_activation
        self.edge_activation = edge_activation

    def build(self, input_shape):
        assert len(input_shape) == 3  # X, A, E, right?
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )

        self.stack_models = []
        self.stack_model_acts = []

        if isinstance(self.stack_channels, Iterable):
            assert len(self.stack_channels) > 0

            for count, value in enumerate(self.stack_channels):
                self.stack_models.append(Dense(value, **layer_kwargs))
                if count != len(self.stack_channels) - 1:
                    self.stack_model_acts.append(ReLU())
                else:
                    self.stack_model_acts.append(PReLU())
        else:
            self.stack_models.append(Dense(self.stack_channels, **layer_kwargs))
            self.stack_model_acts.append(PReLU())

        self.node_model = Dense(
            self.node_channels, activation=self.node_activation, **layer_kwargs
        )
        self.edge_model = Dense(
            self.edge_channels, activation=self.edge_activation, **layer_kwargs
        )

        if self.attention:
            self.incoming_att_sigmoid = Dense(1, activation="sigmoid")
            self.incoming_att_multiply = Multiply()
            self.outgoing_att_sigmoid = Dense(1, activation="sigmoid")
            self.outgoing_att_multiply = Multiply()

        self.built = True

    def call(self, inputs, **kwargs):
        x, a, e = self.get_inputs(inputs)
        x_out, e_out = self.propagate(x, a, e)

        return x_out, e_out

    def message(self, x, e=None):
        x_i = self.get_targets(x)  # Features of self
        x_j = self.get_sources(x)  # Features of neighbours

        # Features of outgoing edges are simply the edge features
        e_ij = e

        # Features of incoming edges j -> i are obtained by transposing the edge features.
        # Since TF does not allow transposing sparse matrices with rank > 2, we instead
        # re-order a tf.range(n_edges) and use it as indices to re-order the edge
        # features.
        # The following two instructions are the sparse equivalent of
        #     tf.transpose(E, perm=(1, 0, 2))
        # where E has shape (N, N, S).
        reorder_idx = gen_sparse_ops.sparse_reorder(
            tf.stack([self.index_targets, self.index_sources], axis=-1),
            tf.range(tf.shape(e)[0]),
            (self.n_nodes, self.n_nodes),
        )[1]
        e_ji = tf.gather(e, reorder_idx)

        # Concatenate the features and feed to first MLP
        stack_ij = tf.concat(
            [x_i, x_j, e_ij, e_ji], axis=-1
        )  # Shape: (n_edges, F + F + S + S)

        for stack_conv in range(0, len(self.stack_models)):
            stack_ij = self.stack_models[stack_conv](stack_ij)
            stack_ij = self.stack_model_acts[stack_conv](stack_ij)

        return stack_ij

    def aggregate(self, messages, x=None):
        if self.attention:
            incoming_att = self.incoming_att_sigmoid(messages)
            incoming = self.incoming_att_multiply([incoming_att, messages])
            incoming = self.agg(incoming, self.index_targets, self.n_nodes)
            outgoing_att = self.outgoing_att_sigmoid(messages)
            outgoing = self.outgoing_att_multiply([outgoing_att, messages])
            outgoing = self.agg(outgoing, self.index_sources, self.n_nodes)
        else:
            incoming = self.agg(messages, self.index_targets, self.n_nodes)
            outgoing = self.agg(messages, self.index_sources, self.n_nodes)

        return tf.concat([x, incoming, outgoing], axis=-1), messages

    def update(self, embeddings):
        x_new, stack_ij = embeddings

        return self.node_model(x_new), self.edge_model(stack_ij)

    @property
    def config(self):
        return {
            "stack_channels": self.stack_channels,
            "node_channels": self.node_channels,
            "edge_channels": self.edge_channels,
            "attention": self.attention,
            "node_activation": self.node_activation,
            "edge_activation": self.edge_activation,
        }


class XENetConvBatch(Conv):
    r"""
    A batch-mode version of XENetConv.

    **Mode**: batch.

    **This layer expects a dense adjacency matrix.**
    """

    def __init__(
        self,
        stack_channels,
        node_channels,
        edge_channels,
        attention: bool = True,
        node_activation=None,
        edge_activation=None,
        aggregate: str = "sum",
        use_bias: bool = True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.stack_channels = stack_channels
        self.node_channels = node_channels
        self.edge_channels = edge_channels
        self.attention = attention
        self.node_activation = node_activation
        self.edge_activation = edge_activation

    def build(self, input_shape):
        assert len(input_shape) == 3  # X, A, E, right?
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
            dtype=self.dtype,
        )
        channels = input_shape[0][-1]  # last dim of X

        self.stack_models = []
        self.stack_model_acts = []

        if isinstance(self.stack_channels, Iterable):
            assert len(self.stack_channels) > 0

            for count, value in enumerate(self.stack_channels):
                self.stack_models.append(Dense(value, **layer_kwargs))
                if count != len(self.stack_channels) - 1:
                    self.stack_model_acts.append(ReLU())
                else:
                    self.stack_model_acts.append(PReLU(shared_axes=[1, 2]))
        else:
            self.stack_models.append(Dense(self.stack_channels, **layer_kwargs))
            self.stack_model_acts.append(PReLU(shared_axes=[1, 2]))

        self.node_model = Dense(
            self.node_channels, activation=self.node_activation, **layer_kwargs
        )
        self.edge_model = Dense(
            self.edge_channels, activation=self.edge_activation, **layer_kwargs
        )

        if self.attention:
            self.incoming_att_sigmoid = Dense(1, activation="sigmoid")
            self.incoming_att_multiply = Multiply()
            self.outgoing_att_sigmoid = Dense(1, activation="sigmoid")
            self.outgoing_att_multiply = Multiply()

        self.built = True

    def call(self, inputs):
        x, a, e = inputs

        assert len(x.shape) == 3
        assert len(e.shape) == 4

        N = tf.shape(x)[1]

        xi_shape = [-1, N, 1, x.shape[2]]
        xj_shape = [-1, 1, N, x.shape[2]]

        xi = tf.reshape(x, xi_shape)  # b n 1 f
        xj = tf.reshape(x, xj_shape)  # b 1 n f

        xi = tf.repeat(xi, N, axis=2)
        xj = tf.repeat(xj, N, axis=1)

        e_transpose = tf.transpose(e, perm=[0, 2, 1, 3])

        stack = tf.concat([xi, xj, e, e_transpose], axis=-1)

        for i in range(0, len(self.stack_models)):
            stack = self.stack_models[i](stack)
            stack = self.stack_model_acts[i](stack)

        e_mask_shape = [-1, tf.shape(a)[1], tf.shape(a)[2], 1]
        e_mask = tf.reshape(a, e_mask_shape)

        # zero-out elements that aren't edges in the adjacency matrix
        stack = Multiply()([stack, e_mask])

        if self.attention:
            att1 = self.incoming_att_sigmoid(stack)
            incoming_e = self.incoming_att_multiply([stack, att1])
            incoming_e = tf.keras.backend.sum(incoming_e, axis=-2, keepdims=False)
            att2 = self.outgoing_att_sigmoid(stack)
            outgoing_e = self.outgoing_att_multiply([stack, att2])
            outgoing_e = tf.keras.backend.sum(outgoing_e, axis=-3, keepdims=False)
        else:
            incoming_e = tf.keras.backend.sum(stack, axis=-2, keepdims=False)
            outgoing_e = tf.keras.backend.sum(stack, axis=-3, keepdims=False)

        final_stack = Concatenate(axis=-1)([x, incoming_e, outgoing_e])

        x = self.node_model(final_stack)
        e = self.edge_model(stack)

        # zero-out elements that aren't edges in the adjacency matrix
        e = Multiply()([e, e_mask])
        return x, e

    @property
    def config(self):
        return {
            "stack_channels": self.stack_channels,
            "node_channels": self.node_channels,
            "edge_channels": self.edge_channels,
            "attention": self.attention,
            "node_activation": self.node_activation,
            "edge_activation": self.edge_activation,
        }



import inspect

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer

from spektral.utils.keras import (
    deserialize_kwarg,
    is_keras_kwarg,
    is_layer_kwarg,
    serialize_kwarg,
)


class SRCPool(Layer):
    r"""
    A general class for graph pooling layers based on the "Select, Reduce,
    Connect" framework presented in:

    > [Understanding Pooling in Graph Neural Networks.](https://arxiv.org/abs/2110.05292)<br>
    > Daniele Grattarola et al.

    This layer computes:
    $$
        \begin{align}
            & \mathcal{S} = \left\{\mathcal{S}_k\right\}_{k=1:K} = \textrm{Sel}(\mathcal{G}) \\
            & \mathcal{X}'=\left\{\textrm{Red}( \mathcal{G}, \mathcal{S}_k )\right\}_{k=1:K} \\
            & \mathcal{E}'=\left\{\textrm{Con}( \mathcal{G}, \mathcal{S}_k, \mathcal{S}_l )\right\}_{k,L=1:K} \\
        \end{align}
    $$
    Where \(\textrm{Sel}\) is a node equivariant selection function that computes
    the supernode assignments \(\mathcal{S}_k\), \(\textrm{Red}\) is a
    permutation-invariant function to reduce the supernodes into the new node
    attributes, and \(\textrm{Con}\) is a permutation-invariant connection
    function that computes the link between the pooled nodes.

    By extending this class, it is possible to create any pooling layer in the
    SRC formalism.

    **Input**

    - `x`: Tensor of shape `([batch], N, F)` representing node features;
    - `a`: Tensor or SparseTensor of shape `([batch], N, N)` representing the
    adjacency matrix;
    - `i`: (optional) Tensor of integers with shape `(N, )` representing the
    batch index;

    **Output**

    - `x_pool`: Tensor of shape `([batch], K, F)`, representing the node
    features of the output. `K` is the number of output nodes and depends on the
    specific pooling strategy;
    - `a_pool`: Tensor or SparseTensor of shape `([batch], K, K)` representing
    the adjacency matrix of the output;
    - `i_pool`: (only if i was given as input) Tensor of integers with shape
    `(K, )` representing the batch index of the output;
    - `s`: (if `return_selection=True`) Tensor or SparseTensor representing the
    supernode assignments;

    **API**

    - `pool(x, a, i, **kwargs)`: pools the graph and returns the reduced node
    features and adjacency matrix. If the batch index `i` is not `None`, a
    reduced version of `i` will be returned as well.
    Any given `kwargs` will be passed as keyword arguments to `select()`,
    `reduce()` and `connect()` if any matching key is found.
    The mandatory arguments of `pool()` **must** be computed in `call()` by
    calling `self.get_inputs(inputs)`.
    - `select(x, a, i, **kwargs)`: computes supernode assignments mapping the
    nodes of the input graph to the nodes of the output.
    - `reduce(x, s, **kwargs)`: reduces the supernodes to form the nodes of the
    pooled graph.
    - `connect(a, s, **kwargs)`: connects the reduced supernodes.
    - `reduce_index(i, s, **kwargs)`: helper function to reduce the batch index
    (only called if `i` is given as input).

    When overriding any function of the API, it is possible to access the
    true number of nodes of the input (`n_nodes`) as a Tensor in the instance variable
    `self.n_nodes` (this is populated by `self.get_inputs()` at the beginning of
    `call()`).

    **Arguments**:

    - `return_selection`: if `True`, the Tensor used to represent supernode assignments
    will be returned with `x_pool`, `a_pool`, and `i_pool`;
    """

    def __init__(self, return_selection=False, **kwargs):
        # kwargs for the Layer class are handled automatically
        super().__init__(**{k: v for k, v in kwargs.items() if is_keras_kwarg(k)})
        self.supports_masking = True
        self.return_selection = return_selection

        # *_regularizer, *_constraint, *_initializer, activation, and use_bias
        # are dealt with automatically if passed to the constructor
        self.kwargs_keys = []
        for key in kwargs:
            if is_layer_kwarg(key):
                attr = kwargs[key]
                attr = deserialize_kwarg(key, attr)
                self.kwargs_keys.append(key)
                setattr(self, key, attr)

        # Signature of the SRC functions
        self.sel_signature = inspect.signature(self.select).parameters
        self.red_signature = inspect.signature(self.reduce).parameters
        self.con_signature = inspect.signature(self.connect).parameters
        self.i_red_signature = inspect.signature(self.reduce_index).parameters

        self._n_nodes = None

    def build(self, input_shape):
        super().build(input_shape)

    def call(self, inputs, **kwargs):
        # Always start the call() method with get_inputs(inputs) to set self.n_nodes
        x, a, i = self.get_inputs(inputs)

        return self.pool(x, a, i, **kwargs)

    def pool(self, x, a, i, **kwargs):
        """
        This is the core method of the SRC class, which runs a full pass of
        selection, reduction and connection.
        It is usually not necessary to modify this function. Any previous/shared
        operations should be done in `call()` and their results can be passed to
        the three SRC functions via keyword arguments (any kwargs given to this
        function will be matched to the signature of `select()`, `reduce()` and
        `connect()` and propagated as input to the three functions).
        Any pooling logic should go in the SRC functions themselves.
        :param x: Tensor of shape `([batch], N, F)`;
        :param a: Tensor or SparseTensor of shape `([batch], N, N)`;
        :param i: only in single/disjoint mode, Tensor of integers with shape
        `(N, )`; otherwise, `None`;
        :param kwargs: additional keyword arguments for `select()`, `reduce()`
        and `connect()`. Any matching kwargs will be passed to each of the three
        functions.
        :return:
            - `x_pool`: Tensor of shape `([batch], K, F)`, where `K` is the
            number of output nodes and depends on the pooling strategy;
            - `a_pool`: Tensor or SparseTensor of shape `([batch], K, K)`;
            - `i_pool`: (only if `i` is not `None`) Tensor of integers with shape
            `(K, )`;
        """
        # Select
        sel_kwargs = self._get_kwargs(x, a, i, self.sel_signature, kwargs)
        s = self.select(x, a, i, **sel_kwargs)

        # Reduce
        red_kwargs = self._get_kwargs(x, a, i, self.red_signature, kwargs)
        x_pool = self.reduce(x, s, **red_kwargs)

        # Index reduce
        i_red_kwargs = self._get_kwargs(x, a, i, self.i_red_signature, kwargs)
        i_pool = self.reduce_index(i, s, **i_red_kwargs) if i is not None else None

        # Connect
        con_kwargs = self._get_kwargs(x, a, i, self.con_signature, kwargs)
        a_pool = self.connect(a, s, **con_kwargs)

        return self.get_outputs(x_pool, a_pool, i_pool, s)

    def select(self, x, a, i, **kwargs):
        """
        Selection function. Given the graph, computes the supernode assignments
        that will eventually be mapped to the `K` nodes of the pooled graph.
        Supernode assignments are usually represented as a dense matrix of shape
        `(N, K)` or sparse indices of shape `(K, )`.
        :param x: Tensor of shape `([batch], N, F)`;
        :param a: Tensor or SparseTensor (depending on the implementation of the
        SRC functions) of shape `([batch], N, N)`;
        :param i: Tensor of integers with shape `(N, )` or `None`;
        :param kwargs: additional keyword arguments.
        :return: Tensor representing supernode assignments.
        """
        return tf.range(tf.shape(i))

    def reduce(self, x, s, **kwargs):
        """
        Reduction function. Given a selection, reduces the supernodes to form
        the nodes of the new graph.
        :param x: Tensor of shape `([batch], N, F)`;
        :param s: Tensor representing supernode assignments, as computed by
        `select()`;
        :param kwargs: additional keyword arguments; when overriding this
        function, any keyword argument defined explicitly as `key=None` will be
        automatically filled in when calling `pool(key=value)`.
        :return: Tensor of shape `([batch], K, F)` representing the node attributes of
        the pooled graph.
        """
        return tf.gather(x, s)

    def connect(self, a, s, **kwargs):
        """
        Connection function. Given a selection, connects the nodes of the pooled
        graphs.
        :param a: Tensor or SparseTensor of shape `([batch], N, N)`;
        :param s: Tensor representing supernode assignments, as computed by
        `select()`;
        :param kwargs: additional keyword arguments; when overriding this
        function, any keyword argument defined explicitly as `key=None` will be
        automatically filled in when calling `pool(key=value)`.
        :return: Tensor or SparseTensor of shape `([batch], K, K)` representing
        the adjacency matrix of the pooled graph.
        """
        return sparse_connect(a, s, self.n_nodes)

    def reduce_index(self, i, s, **kwargs):
        """
        Helper function to reduce the batch index `i`. Given a selection,
        returns a new batch index for the pooled graph. This is only called by
        `pool()` when `i` is given as input to the layer.
        :param i: Tensor of integers with shape `(N, )`;
        :param s: Tensor representing supernode assignments, as computed by
        `select()`.
        :param kwargs: additional keyword arguments; when overriding this
        function, any keyword argument defined explicitly as `key=None` will be
        automatically filled in when calling `pool(key=value)`.
        :return: Tensor of integers of shape `(K, )`.
        """
        return tf.gather(i, s)

    @staticmethod
    def _get_kwargs(x, a, i, signature, kwargs):
        output = {}
        for k in signature.keys():
            if signature[k].default is inspect.Parameter.empty or k == "kwargs":
                pass
            elif k == "x":
                output[k] = x
            elif k == "a":
                output[k] = a
            elif k == "i":
                output[k] = i
            elif k in kwargs:
                output[k] = kwargs[k]
            else:
                raise ValueError("Missing key {} for signature {}".format(k, signature))

        return output

    def get_inputs(self, inputs):
        if len(inputs) == 3:
            x, a, i = inputs
            if K.ndim(i) == 2:
                i = i[:, 0]
            assert K.ndim(i) == 1, "i must have rank 1"
        elif len(inputs) == 2:
            x, a = inputs
            i = None
        else:
            raise ValueError(
                "Expected 2 or 3 inputs tensors (x, a, i), got {}.".format(len(inputs))
            )

        self.n_nodes = tf.shape(x)[-2]

        return x, a, i

    def get_outputs(self, x_pool, a_pool, i_pool, s):
        output = [x_pool, a_pool]
        if i_pool is not None:
            output.append(i_pool)
        if self.return_selection:
            output.append(s)

        return output

    def get_config(self):
        config = {
            "return_selection": self.return_selection,
        }
        for key in self.kwargs_keys:
            config[key] = serialize_kwarg(key, getattr(self, key))
        base_config = super().get_config()
        return {**base_config, **config}

    def compute_mask(self, inputs, mask=None):
        # After pooling all nodes are always valid
        return None

    @property
    def n_nodes(self):
        if self._n_nodes is None:
            raise ValueError(
                "self.n_nodes has not been defined. Have you called "
                "self.get_inputs(inputs) at the beginning of call()?"
            )
        return self._n_nodes

    @n_nodes.setter
    def n_nodes(self, value):
        self._n_nodes = value

    @n_nodes.deleter
    def n_nodes(self):
        self._n_nodes = None


def sparse_connect(A, S, N):
    N_sel = tf.cast(tf.shape(S), tf.int64)[0]
    m = tf.scatter_nd(S[:, None], tf.range(N_sel) + 1, (N,)) - 1

    row, col = A.indices[:, 0], A.indices[:, 1]
    r_mask = tf.gather(m, row)
    c_mask = tf.gather(m, col)
    mask_total = (r_mask >= 0) & (c_mask >= 0)
    r_new = tf.boolean_mask(r_mask, mask_total)
    c_new = tf.boolean_mask(c_mask, mask_total)
    v_new = tf.boolean_mask(A.values, mask_total)

    output = tf.SparseTensor(
        values=v_new, indices=tf.stack((r_new, c_new), 1), dense_shape=(N_sel, N_sel)
    )
    return tf.sparse.reorder(output)



import numpy as np


class Constant:
    """
    Concatenates a constant value to the node attributes.

    **Arguments**

    - `value`: the value to concatenate to the node attributes.
    """

    def __init__(self, value):
        self.value = value

    def __call__(self, graph):
        value = np.zeros((graph.n_nodes, 1)) + self.value
        if graph.x is None:
            graph.x = value
        else:
            graph.x = np.concatenate((graph.x, value), axis=-1)

        return graph



import numpy as np
import tensorflow as tf

from spektral.data.utils import (
    batch_generator,
    collate_labels_batch,
    collate_labels_disjoint,
    get_spec,
    prepend_none,
    sp_matrices_to_sp_tensors,
    to_batch,
    to_disjoint,
    to_mixed,
    to_tf_signature,
)

version = tf.__version__.split(".")
major, minor = int(version[0]), int(version[1])
tf_loader_available = major >= 2 and minor >= 4


class Loader:
    """
    Parent class for data loaders. The role of a Loader is to iterate over a
    Dataset and yield batches of graphs to feed your Keras Models.

    This is achieved by having a generator object that produces lists of Graphs,
    which are then collated together and returned as Tensors.

    The core of a Loader is the `collate(batch)` method.
    This takes as input a list of `Graph` objects and returns a list of Tensors,
    np.arrays, or SparseTensors.

    For instance, if all graphs have the same number of nodes and size of the
    attributes, a simple collation function can be:

    ```python
    def collate(self, batch):
        x = np.array([g.x for g in batch])
        a = np.array([g.a for g in batch)]
        return x, a
    ```

    The `load()` method of a Loader returns an object that can be passed to a Keras
    model when using the `fit`, `predict` and `evaluate` functions.
    You can use it as follows:

    ```python
    model.fit(loader.load(), steps_per_epoch=loader.steps_per_epoch)
    ```

    The `steps_per_epoch` property represents the number of batches that are in
    an epoch, and is a required keyword when calling `fit`, `predict` or `evaluate`
    with a Loader.

    If you are using a custom training function, you can specify the input signature
    of your batches with the tf.TypeSpec system to avoid unnecessary re-tracings.
    The signature is computed automatically by calling `loader.tf_signature()`.

    For example, a simple training step can be written as:

    ```python
    @tf.function(input_signature=loader.tf_signature())  # Specify signature here
    def train_step(inputs, target):
        with tf.GradientTape() as tape:
            predictions = model(inputs, training=True)
            loss = loss_fn(target, predictions)
        gradients = tape.gradient(loss, model.trainable_variables)
        optimizer.apply_gradients(zip(gradients, model.trainable_variables))
    ```

    We can then train our model in a loop as follows:

    ```python
    for batch in loader:
        train_step(*batch)
    ```

    **Arguments**

    - `dataset`: a `spektral.data.Dataset` object;
    - `batch_size`: int, size of the mini-batches;
    - `epochs`: int, number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: bool, whether to shuffle the dataset at the start of each epoch.
    """

    def __init__(self, dataset, batch_size=1, epochs=None, shuffle=True):
        self.dataset = dataset
        self.batch_size = batch_size
        self.epochs = epochs
        self.shuffle = shuffle
        self._generator = self.generator()

    def __iter__(self):
        return self

    def __next__(self):
        nxt = self._generator.__next__()
        return self.collate(nxt)

    def generator(self):
        """
        Returns lists (batches) of `Graph` objects.
        """
        return batch_generator(
            self.dataset,
            batch_size=self.batch_size,
            epochs=self.epochs,
            shuffle=self.shuffle,
        )

    def collate(self, batch):
        """
        Converts a list of graph objects to Tensors or np.arrays representing the batch.
        :param batch: a list of `Graph` objects.
        """
        raise NotImplementedError

    def load(self):
        """
        Returns an object that can be passed to a Keras model when using the `fit`,
        `predict` and `evaluate` functions.
        By default, returns the Loader itself, which is a generator.
        """
        return self

    def tf_signature(self):
        """
        Returns the signature of the collated batches using the tf.TypeSpec system.
        By default, the signature is that of the dataset (`dataset.signature`):

            - Adjacency matrix has shape `[n_nodes, n_nodes]`
            - Node features have shape `[n_nodes, n_node_features]`
            - Edge features have shape `[n_edges, n_node_features]`
            - Targets have shape `[..., n_labels]`
        """
        signature = self.dataset.signature
        return to_tf_signature(signature)

    def pack(self, batch):
        """
        Given a batch of graphs, groups their attributes into separate lists and packs
        them in a dictionary.

        For instance, if a batch has three graphs g1, g2 and g3 with node
        features (x1, x2, x3) and adjacency matrices (a1, a2, a3), this method
        will return a dictionary:

        ```python
        >>> {'a_list': [a1, a2, a3], 'x_list': [x1, x2, x3]}
        ```

        :param batch: a list of `Graph` objects.
        """
        output = [list(elem) for elem in zip(*[g.numpy() for g in batch])]
        keys = [k + "_list" for k in self.dataset.signature.keys()]
        return dict(zip(keys, output))

    @property
    def steps_per_epoch(self):
        """
        :return: the number of batches of size `self.batch_size` in the dataset (i.e.,
        how many batches are in an epoch).
        """
        return int(np.ceil(len(self.dataset) / self.batch_size))


class SingleLoader(Loader):
    """
    A Loader for [single mode](https://graphneural.network/data-modes/#single-mode).

    This loader produces Tensors representing a single graph. As such, it can
    only be used with Datasets of length 1 and the `batch_size` cannot be set.

    The loader supports sample weights through the `sample_weights` argument.
    If given, then each batch will be a tuple `(inputs, labels, sample_weights)`.

    **Arguments**

    - `dataset`: a `spektral.data.Dataset` object with only one graph;
    - `epochs`: int, number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: bool, whether to shuffle the data at the start of each epoch;
    - `sample_weights`: Numpy array, will be appended to the output automatically.

    **Output**

    Returns a tuple `(inputs, labels)` or `(inputs, labels, sample_weights)`.

    `inputs` is a tuple containing the data matrices of the graph, only if they
    are not `None`:

    - `x`: same as `dataset[0].x`;
    - `a`: same as `dataset[0].a` (scipy sparse matrices are converted to
    SparseTensors);
    - `e`: same as `dataset[0].e`;

    `labels` is the same as `dataset[0].y`.

    `sample_weights` is the same array passed when creating the loader.
    """

    def __init__(self, dataset, epochs=None, sample_weights=None):
        if len(dataset) != 1:
            raise ValueError(
                "SingleLoader can only be used with Datasets that"
                "have a single graph."
            )
        self.sample_weights = sample_weights
        super().__init__(dataset, batch_size=1, epochs=epochs, shuffle=False)

    def collate(self, batch):
        packed = self.pack(batch)

        y = packed.pop("y_list", None)
        if y is not None:
            y = collate_labels_disjoint(y, node_level=True)

        output = to_disjoint(**packed)
        output = output[:-1]  # Discard batch index
        output = sp_matrices_to_sp_tensors(output)

        if len(output) == 1:
            output = output[0]

        output = (output,)
        if y is not None:
            output += (y,)
        if self.sample_weights is not None:
            output += (self.sample_weights,)

        if len(output) == 1:
            output = output[0]  # Again, in case there are no targets and no SW

        return output

    def load(self):
        output = self.collate(self.dataset)
        return tf.data.Dataset.from_tensors(output).repeat(self.epochs)


class DisjointLoader(Loader):
    """
    A Loader for [disjoint mode](https://graphneural.network/data-modes/#disjoint-mode).

    This loader represents a batch of graphs via their disjoint union.

    The loader automatically computes a batch index tensor, containing integer
    indices that map each node to its corresponding graph in the batch.

    The adjacency matrix os returned as a SparseTensor, regardless of the input.

    If `node_level=False`, the labels are interpreted as graph-level labels and
    are stacked along an additional dimension.
    If `node_level=True`, then the labels are stacked vertically.

    **Note:** TensorFlow 2.4 or above is required to use this Loader's `load()`
    method in a Keras training loop.

    **Arguments**

    - `dataset`: a graph Dataset;
    - `node_level`: bool, if `True` stack the labels vertically for node-level
    prediction;
    - `batch_size`: size of the mini-batches;
    - `epochs`: number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: whether to shuffle the data at the start of each epoch.

    **Output**

    For each batch, returns a tuple `(inputs, labels)`.

    `inputs` is a tuple containing:

    - `x`: node attributes of shape `[n_nodes, n_node_features]`;
    - `a`: adjacency matrices of shape `[n_nodes, n_nodes]`;
    - `e`: edge attributes of shape `[n_edges, n_edge_features]`;
    - `i`: batch index of shape `[n_nodes]`.

    `labels` have shape `[batch, n_labels]` if `node_level=False` or
    `[n_nodes, n_labels]` otherwise.

    """

    def __init__(
        self, dataset, node_level=False, batch_size=1, epochs=None, shuffle=True
    ):
        self.node_level = node_level
        super().__init__(dataset, batch_size=batch_size, epochs=epochs, shuffle=shuffle)

    def collate(self, batch):
        packed = self.pack(batch)

        y = packed.pop("y_list", None)
        if y is not None:
            y = collate_labels_disjoint(y, node_level=self.node_level)

        output = to_disjoint(**packed)
        output = sp_matrices_to_sp_tensors(output)

        if len(output) == 1:
            output = output[0]

        if y is None:
            return output
        else:
            return output, y

    def load(self):
        if not tf_loader_available:
            raise RuntimeError(
                "Calling DisjointLoader.load() requires " "TensorFlow 2.4 or greater."
            )
        return tf.data.Dataset.from_generator(
            lambda: self, output_signature=self.tf_signature()
        )

    def tf_signature(self):
        """
        Adjacency matrix has shape [n_nodes, n_nodes]
        Node features have shape [n_nodes, n_node_features]
        Edge features have shape [n_edges, n_edge_features]
        Targets have shape [*, n_labels]
        """
        signature = self.dataset.signature
        if "y" in signature:
            signature["y"]["shape"] = prepend_none(signature["y"]["shape"])
        if "a" in signature:
            signature["a"]["spec"] = tf.SparseTensorSpec

        signature["i"] = dict()
        signature["i"]["spec"] = tf.TensorSpec
        signature["i"]["shape"] = (None,)
        signature["i"]["dtype"] = tf.as_dtype(tf.int64)

        return to_tf_signature(signature)


class BatchLoader(Loader):
    """
    A Loader for [batch mode](https://graphneural.network/data-modes/#batch-mode).

    This loader returns batches of graphs stacked along an extra dimension,
    with all "node" dimensions padded to be equal among all graphs.

    If `n_max` is the number of nodes of the biggest graph in the batch, then
    the padding consist of adding zeros to the node features, adjacency matrix,
    and edge attributes of each graph so that they have shapes
    `[n_max, n_node_features]`, `[n_max, n_max]`, and
    `[n_max, n_max, n_edge_features]` respectively.

    The zero-padding is done batch-wise, which saves up memory at the cost of
    more computation. If latency is an issue but memory isn't, or if the
    dataset has graphs with a similar number of nodes, you can use
    the `PackedBatchLoader` that zero-pads all the dataset once and then
    iterates over it.

    Note that the adjacency matrix and edge attributes are returned as dense
    arrays.

    if `mask=True`, node attributes will be extended with a binary mask that indicates
    valid nodes (the last feature of each node will be 1 if the node was originally in
    the graph and 0 if it is a fake node added by zero-padding).

    Use this flag in conjunction with layers.base.GraphMasking to start the propagation
    of masks in a model (necessary for node-level prediction and models that use a
    dense pooling layer like DiffPool or MinCutPool).

    If `node_level=False`, the labels are interpreted as graph-level labels and
    are returned as an array of shape `[batch, n_labels]`.
    If `node_level=True`, then the labels are padded along the node dimension and are
    returned as an array of shape `[batch, n_max, n_labels]`.

    **Arguments**

    - `dataset`: a graph Dataset;
    - `mask`: bool, whether to add a mask to the node features;
    - `batch_size`: int, size of the mini-batches;
    - `epochs`: int, number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: bool, whether to shuffle the data at the start of each epoch;
    - `node_level`: bool, if `True` pad the labels along the node dimension;

    **Output**

    For each batch, returns a tuple `(inputs, labels)`.

    `inputs` is a tuple containing:

    - `x`: node attributes of shape `[batch, n_max, n_node_features]`;
    - `a`: adjacency matrices of shape `[batch, n_max, n_max]`;
    - `e`: edge attributes of shape `[batch, n_max, n_max, n_edge_features]`.

    `labels` have shape `[batch, n_labels]` if `node_level=False` or
    `[batch, n_max, n_labels]` otherwise.
    """

    def __init__(
        self,
        dataset,
        mask=False,
        batch_size=1,
        epochs=None,
        shuffle=True,
        node_level=False,
    ):
        self.mask = mask
        self.node_level = node_level
        self.signature = dataset.signature
        super().__init__(dataset, batch_size=batch_size, epochs=epochs, shuffle=shuffle)

    def collate(self, batch):
        packed = self.pack(batch)

        y = packed.pop("y_list", None)
        if y is not None:
            y = collate_labels_batch(y, node_level=self.node_level)

        output = to_batch(**packed, mask=self.mask)
        output = sp_matrices_to_sp_tensors(output)

        if len(output) == 1:
            output = output[0]

        if y is None:
            return output
        else:
            return output, y

    def tf_signature(self):
        """
        Adjacency matrix has shape [batch, n_nodes, n_nodes]
        Node features have shape [batch, n_nodes, n_node_features]
        Edge features have shape [batch, n_nodes, n_nodes, n_edge_features]
        Labels have shape [batch, n_labels]
        """
        signature = self.signature
        for k in signature:
            signature[k]["shape"] = prepend_none(signature[k]["shape"])
        if "x" in signature and self.mask:
            # In case we have a mask, the mask is concatenated to the features
            signature["x"]["shape"] = signature["x"]["shape"][:-1] + (
                signature["x"]["shape"][-1] + 1,
            )
        if "a" in signature:
            # Adjacency matrix in batch mode is dense
            signature["a"]["spec"] = tf.TensorSpec
        if "e" in signature:
            # Edge attributes have an extra None dimension in batch mode
            signature["e"]["shape"] = prepend_none(signature["e"]["shape"])
        if "y" in signature and self.node_level:
            # Node labels have an extra None dimension
            signature["y"]["shape"] = prepend_none(signature["y"]["shape"])

        return to_tf_signature(signature)


class PackedBatchLoader(BatchLoader):
    """
    A `BatchLoader` that zero-pads the graphs before iterating over the dataset.
    This means that `n_max` is computed over the whole dataset and not just
    a single batch.

    While using more memory than `BatchLoader`, this loader should reduce the
    computational overhead of padding each batch independently.

    Use this loader if:

    - memory usage isn't an issue and you want to produce the batches as fast
    as possible;
    - the graphs in the dataset have similar sizes and there are no outliers in
    the dataset (i.e., anomalous graphs with many more nodes than the dataset
    average).

    **Arguments**

    - `dataset`: a graph Dataset;
    - `mask`: bool, whether to add a mask to the node features;
    - `batch_size`: int, size of the mini-batches;
    - `epochs`: int, number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: bool, whether to shuffle the data at the start of each epoch;
    - `node_level`: bool, if `True` pad the labels along the node dimension;

    **Output**

    For each batch, returns a tuple `(inputs, labels)`.

    `inputs` is a tuple containing:

    - `x`: node attributes of shape `[batch, n_max, n_node_features]`;
    - `a`: adjacency matrices of shape `[batch, n_max, n_max]`;
    - `e`: edge attributes of shape `[batch, n_max, n_max, n_edge_features]`.

    `labels` have shape `[batch, n_labels]` if `node_level=False` or
    `[batch, n_max, n_labels]` otherwise.
    """

    def __init__(
        self,
        dataset,
        mask=False,
        batch_size=1,
        epochs=None,
        shuffle=True,
        node_level=False,
    ):
        super().__init__(
            dataset,
            mask=mask,
            batch_size=batch_size,
            epochs=epochs,
            shuffle=shuffle,
            node_level=node_level,
        )

        # Drop the Dataset container and work on packed tensors directly
        packed = self.pack(self.dataset)

        y = packed.pop("y_list", None)
        if y is not None:
            y = collate_labels_batch(y, node_level=self.node_level)

        self.dataset = to_batch(**packed, mask=mask)
        if y is not None:
            self.dataset += (y,)

        # Re-instantiate generator after packing dataset
        self._generator = self.generator()

    def collate(self, batch):
        if len(batch) == 2:
            # If there is only one input, i.e., batch = [x, y], we unpack it
            # like this because Keras does not support input lists with only
            # one tensor.
            return batch[0], batch[1]
        else:
            return batch[:-1], batch[-1]

    @property
    def steps_per_epoch(self):
        if len(self.dataset) > 0:
            return int(np.ceil(len(self.dataset[0]) / self.batch_size))


class MixedLoader(Loader):
    """
    A Loader for [mixed mode](https://graphneural.network/data-modes/#mixed-mode).

    This loader returns batches where the node and edge attributes are stacked
    along an extra dimension, but the adjacency matrix is shared by all graphs.

    The loader expects all node and edge features to have the same number of
    nodes and edges.
    The dataset is pre-packed like in a PackedBatchLoader.

    **Arguments**

    - `dataset`: a graph Dataset;
    - `batch_size`: int, size of the mini-batches;
    - `epochs`: int, number of epochs to iterate over the dataset. By default (`None`)
    iterates indefinitely;
    - `shuffle`: bool, whether to shuffle the data at the start of each epoch.

    **Output**

    For each batch, returns a tuple `(inputs, labels)`.

    `inputs` is a tuple containing:

    - `x`: node attributes of shape `[batch, n_nodes, n_node_features]`;
    - `a`: adjacency matrix of shape `[n_nodes, n_nodes]`;
    - `e`: edge attributes of shape `[batch, n_edges, n_edge_features]`.

    `labels` have shape `[batch, ..., n_labels]`.

    """

    def __init__(self, dataset, batch_size=1, epochs=None, shuffle=True):
        assert dataset.a is not None, (
            "Dataset must be in mixed mode, with only "
            "one adjacency matrix stored in the "
            "dataset's `a` attribute.\n"
            "If your dataset does not have an "
            "adjacency matrix, you can use a "
            "BatchLoader or PackedBatchLoader instead."
        )
        assert "a" not in dataset.signature, (
            "Datasets in mixed mode should not"
            "have the adjacency matrix stored"
            "in their Graph objects."
        )
        super().__init__(dataset, batch_size=batch_size, epochs=epochs, shuffle=shuffle)

    def collate(self, batch):
        packed = self.pack(batch)
        packed["a"] = self.dataset.a

        y = packed.pop("y_list", None)
        if y is not None:
            y = np.array(y)

        output = to_mixed(**packed)
        output = sp_matrices_to_sp_tensors(output)

        if len(output) == 1:
            output = output[0]

        if y is None:
            return output
        else:
            return output, y

    def tf_signature(self):
        """
        Adjacency matrix has shape [n_nodes, n_nodes]
        Node features have shape [batch, n_nodes, n_node_features]
        Edge features have shape [batch, n_edges, n_edge_features]
        Targets have shape [batch, ..., n_labels]
        """
        signature = self.dataset.signature
        for k in ["x", "e", "y"]:
            if k in signature:
                signature[k]["shape"] = prepend_none(signature[k]["shape"])

        signature["a"] = dict()
        signature["a"]["spec"] = get_spec(self.dataset.a)
        signature["a"]["shape"] = (None, None)
        signature["a"]["dtype"] = tf.as_dtype(self.dataset.a.dtype)

        return to_tf_signature(signature)



import numpy as np
import tensorflow as tf
from scipy import sparse as sp


def reorder(edge_index, edge_weight=None, edge_features=None):
    """
    Reorders `edge_index`, `edge_weight`, and `edge_features` according to the row-major
    ordering of `edge_index`.
    :param edge_index: np.array of shape `[n_edges, 2]` (representing rows and columns
    of the adjacency matrix), indices to sort in row-major order.
    :param edge_weight: np.array or None, edge weight to sort according to the new
    order of the indices.
    :param edge_features: np.array or None, edge features to sort according to the new
    order of the indices.
    :return:
        - edge_index: np.array, edge_index sorted in row-major order.
        - edge_weight: If edge_weight is not None, edge_weight sorted according to the
        new order of the indices. Otherwise, None.
        - edge_features: If edge_features is not None, edge_features sorted according to
        the new order of the indices. Otherwise, None.
    """
    sort_idx = np.lexsort(np.flipud(edge_index.T))
    output = [edge_index[sort_idx]]
    if edge_weight is not None:
        output.append(edge_weight[sort_idx])
    if edge_features is not None:
        output.append(edge_features[sort_idx])

    return tuple(output)


def edge_index_to_matrix(edge_index, edge_weight, edge_features=None, shape=None):
    reordered = reorder(edge_index, edge_weight, edge_features)
    a = sp.csr_matrix((reordered[1], reordered[0].T), shape=shape)

    if edge_features is not None:
        return a, reordered[2]
    else:
        return a


def sp_matrix_to_sp_tensor(x):
    """
    Converts a Scipy sparse matrix to a SparseTensor.
    The indices of the output are reordered in the canonical row-major ordering, and
    duplicate entries are summed together (which is the default behaviour of Scipy).

    :param x: a Scipy sparse matrix.
    :return: a SparseTensor.
    """
    if len(x.shape) != 2:
        raise ValueError("x must have rank 2")
    row, col, values = sp.find(x)
    out = tf.SparseTensor(
        indices=np.array([row, col]).T, values=values, dense_shape=x.shape
    )
    return tf.sparse.reorder(out)


def sp_batch_to_sp_tensor(a_list):
    """
    Converts a list of Scipy sparse matrices to a rank 3 SparseTensor.
    :param a_list: list of Scipy sparse matrices with the same shape.
    :return: SparseTensor of rank 3.
    """
    tensor_data = []
    for i, a in enumerate(a_list):
        row, col, values = sp.find(a)
        batch = np.ones_like(col) * i
        tensor_data.append((values, batch, row, col))
    tensor_data = list(map(np.concatenate, zip(*tensor_data)))

    out = tf.SparseTensor(
        indices=np.array(tensor_data[1:]).T,
        values=tensor_data[0],
        dense_shape=(len(a_list),) + a_list[0].shape,
    )

    return out



import networkx as nx
import numpy as np


class ClusteringCoeff:
    """
    Concatenates to each node attribute the clustering coefficient of the
    corresponding node.
    """

    def __call__(self, graph):
        if "a" not in graph:
            raise ValueError("The graph must have an adjacency matrix")
        clustering_coeff = nx.clustering(nx.Graph(graph.a))
        clustering_coeff = np.array(
            [clustering_coeff[i] for i in range(graph.n_nodes)]
        )[:, None]

        if "x" not in graph:
            graph.x = clustering_coeff
        else:
            graph.x = np.concatenate((graph.x, clustering_coeff), axis=-1)

        return graph



import os
import os.path as osp

import numpy as np
from joblib import Parallel, delayed
from tensorflow.keras.utils import get_file
from tqdm import tqdm

from spektral.data import Dataset, Graph
from spektral.utils import label_to_one_hot, sparse
from spektral.utils.io import load_csv, load_sdf

ATOM_TYPES = [1, 6, 7, 8, 9]
BOND_TYPES = [1, 2, 3, 4]


class QM9(Dataset):
    """
    The QM9 chemical data set of small molecules.

    In this dataset, nodes represent atoms and edges represent chemical bonds.
    There are 5 possible atom types (H, C, N, O, F) and 4 bond types (single,
    double, triple, aromatic).

    Node features represent the chemical properties of each atom and include:

    - The atomic number, one-hot encoded;
    - The atom's position in the X, Y, and Z dimensions;
    - The atomic charge;
    - The mass difference from the monoisotope;

    The edge features represent the type of chemical bond between two atoms,
    one-hot encoded.

    Each graph has an 19-dimensional label for regression.

    **Arguments**

    - `amount`: int, load this many molecules instead of the full dataset
    (useful for debugging).
    - `n_jobs`: number of CPU cores to use for reading the data (-1, to use all
    available cores).
    """

    url = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/gdb9.tar.gz"

    def __init__(self, amount=None, n_jobs=1, **kwargs):
        self.amount = amount
        self.n_jobs = n_jobs
        super().__init__(**kwargs)

    def download(self):
        get_file(
            "qm9.tar.gz",
            self.url,
            extract=True,
            cache_dir=self.path,
            cache_subdir=self.path,
        )
        os.remove(osp.join(self.path, "qm9.tar.gz"))

    def read(self):
        print("Loading QM9 dataset.")
        sdf_file = osp.join(self.path, "gdb9.sdf")
        data = load_sdf(sdf_file, amount=self.amount)  # Internal SDF format

        def read_mol(mol):
            x = np.array([atom_to_feature(atom) for atom in mol["atoms"]])
            a, e = mol_to_adj(mol)
            return x, a, e

        data = Parallel(n_jobs=self.n_jobs)(
            delayed(read_mol)(mol) for mol in tqdm(data, ncols=80)
        )
        x_list, a_list, e_list = list(zip(*data))

        # Load labels
        labels_file = osp.join(self.path, "gdb9.sdf.csv")
        labels = load_csv(labels_file)
        labels = labels.set_index("mol_id").values
        if self.amount is not None:
            labels = labels[: self.amount]

        return [
            Graph(x=x, a=a, e=e, y=y)
            for x, a, e, y in zip(x_list, a_list, e_list, labels)
        ]


def atom_to_feature(atom):
    atomic_num = label_to_one_hot(atom["atomic_num"], ATOM_TYPES)
    coords = atom["coords"]
    charge = atom["charge"]
    iso = atom["iso"]

    return np.concatenate((atomic_num, coords, [charge, iso]), -1)


def mol_to_adj(mol):
    row, col, edge_features = [], [], []
    for bond in mol["bonds"]:
        start, end = bond["start_atom"], bond["end_atom"]
        row += [start, end]
        col += [end, start]
        edge_features += [bond["type"]] * 2

    a, e = sparse.edge_index_to_matrix(
        edge_index=np.array((row, col)).T,
        edge_weight=np.ones_like(row),
        edge_features=label_to_one_hot(edge_features, BOND_TYPES),
    )

    return a, e



from tensorflow.keras import activations, constraints, initializers, regularizers

LAYER_KWARGS = {"activation", "use_bias"}
KERAS_KWARGS = {
    "trainable",
    "name",
    "dtype",
    "dynamic",
    "input_dim",
    "input_shape",
    "batch_input_shape",
    "batch_size",
    "weights",
    "activity_regularizer",
    "autocast",
    "implementation",
}


def is_layer_kwarg(key):
    return key not in KERAS_KWARGS and (
        key.endswith("_initializer")
        or key.endswith("_regularizer")
        or key.endswith("_constraint")
        or key in LAYER_KWARGS
    )


def is_keras_kwarg(key):
    return key in KERAS_KWARGS


def deserialize_kwarg(key, attr):
    if key.endswith("_initializer"):
        return initializers.get(attr)
    if key.endswith("_regularizer"):
        return regularizers.get(attr)
    if key.endswith("_constraint"):
        return constraints.get(attr)
    if key == "activation":
        return activations.get(attr)
    return attr


def serialize_kwarg(key, attr):
    if key.endswith("_initializer"):
        return initializers.serialize(attr)
    if key.endswith("_regularizer"):
        return regularizers.serialize(attr)
    if key.endswith("_constraint"):
        return constraints.serialize(attr)
    if key == "activation":
        return activations.serialize(attr)
    if key == "use_bias":
        return attr



import numpy as np


class NormalizeSphere:
    r"""
        Normalizes the node attributes so that they are centered at the origin and
        contained within a sphere of radius 1:
    $$
            \X_{i} \leftarrow \frac{\X_{i} - \bar\X}{\max_{i,j} \X_{ij}}
        $$

        where \( \bar\X \) is the centroid of the node features.
    """

    def __call__(self, graph):
        offset = np.mean(graph.x, -2, keepdims=True)
        scale = np.abs(graph.x).max()
        graph.x = (graph.x - offset) / scale

        return graph



import numpy as np

from spektral.data import Dataset, Graph
from spektral.utils import sparse


class OGB(Dataset):
    """
    Wrapper for datasets from the [Open Graph Benchmark (OGB)](https://ogb.stanford.edu/).

    **Arguments**

    - `dataset`: an OGB library-agnostic dataset.

    """

    def __init__(self, dataset, **kwargs):
        self.dataset = dataset
        super().__init__(**kwargs)

    def read(self):
        if len(self.dataset) > 1:
            return [Graph(*_elem_to_numpy(elem)) for elem in self.dataset]
        else:
            # OGB crashed if we try to iterate over a NodePropPredDataset
            return [Graph(*_elem_to_numpy(self.dataset[0]))]


def _elem_to_numpy(elem):
    graph, label = elem
    n = graph["num_nodes"]
    x = graph["node_feat"]
    row, col = graph["edge_index"]
    e = graph["edge_feat"]
    a_e = sparse.edge_index_to_matrix(
        edge_index=np.array((row, col)).T,
        edge_weight=np.ones_like(row),
        edge_features=e,
        shape=(n, n),
    )
    if e is None:
        a = a_e
    else:
        a, e = a_e

    return x, a, e, label



import tensorflow as tf
from tensorflow.keras import activations
from tensorflow.keras.layers import BatchNormalization, Dense
from tensorflow.keras.models import Sequential

from spektral.layers import ops
from spektral.layers.convolutional.message_passing import MessagePassing


class GINConv(MessagePassing):
    r"""
    A Graph Isomorphism Network (GIN) from the paper

    > [How Powerful are Graph Neural Networks?](https://arxiv.org/abs/1810.00826)<br>
    > Keyulu Xu et al.

    **Mode**: single, disjoint, mixed.

    **This layer expects a sparse adjacency matrix.**

    This layer computes for each node \(i\):
    $$
        \x_i' = \textrm{MLP}\big( (1 + \epsilon) \cdot \x_i + \sum\limits_{j
        \in \mathcal{N}(i)} \x_j \big)
    $$
    where \(\textrm{MLP}\) is a multi-layer perceptron.

    **Input**

    - Node features of shape `(n_nodes, n_node_features)`;
    - Binary adjacency matrix of shape `(n_nodes, n_nodes)`.

    **Output**

    - Node features with the same shape of the input, but the last dimension
    changed to `channels`.

    **Arguments**

    - `channels`: integer, number of output channels;
    - `epsilon`: unnamed parameter, see the original paper and the equation
    above.
    By setting `epsilon=None`, the parameter will be learned (default behaviour).
    If given as a value, the parameter will stay fixed.
    - `mlp_hidden`: list of integers, number of hidden units for each hidden
    layer in the MLP (if None, the MLP has only the output layer);
    - `mlp_activation`: activation for the MLP layers;
    - `mlp_batchnorm`: apply batch normalization after every hidden layer of the MLP;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        epsilon=None,
        mlp_hidden=None,
        mlp_activation="relu",
        mlp_batchnorm=True,
        aggregate="sum",
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            aggregate=aggregate,
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.epsilon = epsilon
        self.mlp_hidden = mlp_hidden if mlp_hidden else []
        self.mlp_activation = activations.get(mlp_activation)
        self.mlp_batchnorm = mlp_batchnorm

    def build(self, input_shape):
        assert len(input_shape) >= 2
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )

        self.mlp = Sequential()
        for channels in self.mlp_hidden:
            self.mlp.add(Dense(channels, self.mlp_activation, **layer_kwargs))
            if self.mlp_batchnorm:
                self.mlp.add(BatchNormalization())
        self.mlp.add(
            Dense(
                self.channels, self.activation, use_bias=self.use_bias, **layer_kwargs
            )
        )

        if self.epsilon is None:
            self.eps = self.add_weight(shape=(1,), initializer="zeros", name="eps")
        else:
            # If epsilon is given, keep it constant
            self.eps = tf.cast(self.epsilon, self.dtype)
        self.one = tf.cast(1, self.dtype)

        self.built = True

    def call(self, inputs, **kwargs):
        x, a, _ = self.get_inputs(inputs)
        output = self.mlp((self.one + self.eps) * x + self.propagate(x, a))

        return output

    @property
    def config(self):
        return {
            "channels": self.channels,
            "epsilon": self.epsilon,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": self.mlp_activation,
            "mlp_batchnorm": self.mlp_batchnorm,
        }


class GINConvBatch(GINConv):
    r"""
    A batch-mode version of GINConv.

    **Mode**: batch.

    **This layer expects a dense adjacency matrix.**
    """

    def call(self, inputs, **kwargs):
        x, a = inputs
        output = self.mlp((self.one + self.eps) * x + ops.modal_dot(a, x))

        return output



import tensorflow as tf
from tensorflow.keras import backend as K

SINGLE = 1  # Single mode    rank(x) = 2, rank(a) = 2
DISJOINT = SINGLE  # Disjoint mode  rank(x) = 2, rank(a) = 2
BATCH = 3  # Batch mode     rank(x) = 3, rank(a) = 3
MIXED = 4  # Mixed mode     rank(x) = 3, rank(a) = 2


def disjoint_signal_to_batch(X, I):
    """
    Converts a disjoint graph signal to batch node by zero-padding.

    :param X: Tensor, node features of shape (nodes, features).
    :param I: Tensor, graph IDs of shape `(n_nodes, )`;
    :return batch: Tensor, batched node features of shape (batch, N_max, n_node_features)
    """
    I = tf.cast(I, tf.int32)
    num_nodes = tf.math.segment_sum(tf.ones_like(I), I)
    start_index = tf.cumsum(num_nodes, exclusive=True)
    n_graphs = tf.shape(num_nodes)[0]
    max_n_nodes = tf.reduce_max(num_nodes)
    batch_n_nodes = tf.shape(I)[0]
    feature_dim = tf.shape(X)[-1]

    index = tf.range(batch_n_nodes)
    index = (index - tf.gather(start_index, I)) + (I * max_n_nodes)
    dense = tf.zeros((n_graphs * max_n_nodes, feature_dim), dtype=X.dtype)
    dense = tf.tensor_scatter_nd_update(dense, index[..., None], X)

    batch = tf.reshape(dense, (n_graphs, max_n_nodes, feature_dim))

    return batch


def disjoint_adjacency_to_batch(A, I):
    """
    Converts a disjoint adjacency matrix to batch node by zero-padding.

    :param A: Tensor, binary adjacency matrix of shape `(n_nodes, n_nodes)`;
    :param I: Tensor, graph IDs of shape `(n_nodes, )`;
    :return: Tensor, batched adjacency matrix of shape `(batch, N_max, N_max)`;
    """
    I = tf.cast(I, tf.int64)
    indices = A.indices
    values = A.values
    i_nodes, j_nodes = indices[:, 0], indices[:, 1]

    graph_sizes = tf.math.segment_sum(tf.ones_like(I), I)
    max_n_nodes = tf.reduce_max(graph_sizes)
    n_graphs = tf.shape(graph_sizes)[0]

    offset = tf.gather(I, i_nodes)
    offset = tf.gather(tf.cumsum(graph_sizes, exclusive=True), offset)

    relative_j_nodes = j_nodes - offset
    relative_i_nodes = i_nodes - offset

    spaced_i_nodes = tf.gather(I, i_nodes) * max_n_nodes + relative_i_nodes
    new_indices = tf.transpose(tf.stack([spaced_i_nodes, relative_j_nodes]))

    n_graphs = tf.cast(n_graphs, new_indices.dtype)
    max_n_nodes = tf.cast(max_n_nodes, new_indices.dtype)

    dense_adjacency = tf.scatter_nd(
        new_indices, values, (n_graphs * max_n_nodes, max_n_nodes)
    )
    batch = tf.reshape(dense_adjacency, (n_graphs, max_n_nodes, max_n_nodes))

    return batch


def autodetect_mode(x, a):
    """
    Returns a code that identifies the data mode from the given node features
    and adjacency matrix(s).
    The output of this function can be used as follows:

    ```py
    from spektral.layers.ops import modes
    mode = modes.autodetect_mode(x, a)
    if mode == modes.SINGLE:
        print('Single!')
    elif mode == modes.BATCH:
        print('Batch!')
    elif mode == modes.MIXED:
        print('Mixed!')
    ```

    :param x: Tensor or SparseTensor representing the node features
    :param a: Tensor or SparseTensor representing the adjacency matrix(s)
    :return: mode of operation as an integer code.
    """
    x_ndim = K.ndim(x)
    a_ndim = K.ndim(a)
    if x_ndim == 2 and a_ndim == 2:
        return SINGLE
    elif x_ndim == 3 and a_ndim == 3:
        return BATCH
    elif x_ndim == 3 and a_ndim == 2:
        return MIXED
    else:
        raise ValueError(
            "Unknown mode for inputs x, a with ranks {} and {}"
            "respectively.".format(x_ndim, a_ndim)
        )



import os
import os.path as osp
import zipfile

import requests
from tqdm import tqdm

_dataset_folder = os.path.join("~", "spektral", "datasets")
_config_path = osp.expanduser(os.path.join("~", "spektral", "config.json"))
if osp.isfile(_config_path):
    import json

    with open(_config_path) as fh:
        _config = json.load(fh)
        _dataset_folder = _config.get("dataset_folder", _dataset_folder)
DATASET_FOLDER = osp.expanduser(_dataset_folder)


def download_file(url, datadir, fname, progress=True, extract=True):
    with requests.get(url, stream=progress) as r:
        r.raise_for_status()
        os.makedirs(datadir, exist_ok=True)
        outfile = osp.join(datadir, fname)
        with open(outfile, "wb") as of:
            if progress:
                pbar = tqdm(
                    total=int(r.headers["Content-Length"]),
                    ncols=80,
                    unit="B",
                    unit_scale=True,
                    unit_divisor=1024,
                )
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk is not None:
                        of.write(chunk)
                        pbar.update(len(chunk))
            else:
                of.write(r.content)

    if extract and fname.endswith(".zip"):
        with zipfile.ZipFile(outfile, "r") as of:
            of.extractall(datadir)
        os.remove(outfile)



import json
import os
import os.path as osp
import shutil

import numpy as np
import scipy.sparse as sp
from networkx.readwrite import json_graph

from spektral.data import Dataset, Graph
from spektral.data.dataset import DATASET_FOLDER
from spektral.datasets.utils import download_file


class GraphSage(Dataset):
    """
    The datasets used in the paper

    > [Inductive Representation Learning on Large Graphs](https://arxiv.org/abs/1706.02216)<br>
    > William L. Hamilton et al.

    The PPI dataset (originally
    [Stark et al. (2006)](https://www.ncbi.nlm.nih.gov/pubmed/16381927))
    for inductive node classification uses positional gene sets, motif gene sets
    and immunological signatures as features and gene ontology sets as labels.

    The Reddit dataset consists of a graph made of Reddit posts in the month of
    September, 2014. The label for each node is the community that a
    post belongs to. The graph is built by sampling 50 large communities and
    two nodes are connected if the same user commented on both. Node features
    are obtained by concatenating the average GloVe CommonCrawl vectors of
    the title and comments, the post's score and the number of comments.

    The train, test, and validation splits are given as binary masks and are
    accessible via the `mask_tr`, `mask_va`, and `mask_te` attributes.

    **Arguments**

    - `name`: name of the dataset to load (`'ppi'`, or `'reddit'`);
    """

    # TODO normalize features?
    # # # Z-score on features (optional)
    # if normalize_features:
    #     from sklearn.preprocessing import StandardScaler
    #     train_ids = np.array([id_map[n] for n in G.nodes()
    #                           if not G.nodes[n]['val'] and not G.nodes[n]['test']])
    #     x_tr = x[train_ids]
    #     scaler = StandardScaler()
    #     scaler.fit(x_tr)
    #     x = scaler.transform(x)

    url = "http://snap.stanford.edu/graphsage/{}.zip"

    def __init__(self, name, **kwargs):
        if name.lower() not in self.available_datasets():
            raise ValueError(
                "Unknown dataset {}. See {}.available_datasets() for a complete list of"
                "available datasets.".format(name, self.__class__.__name__)
            )
        self.name = name.lower()
        self.mask_tr = self.mask_va = self.mask_te = None
        super().__init__(**kwargs)

    @property
    def path(self):
        return osp.join(DATASET_FOLDER, "GraphSage", self.name)

    def read(self):
        npz_file = osp.join(self.path, self.name) + ".npz"
        data = np.load(npz_file)
        x = data["x"]
        a = sp.csr_matrix(
            (data["adj_data"], (data["adj_row"], data["adj_col"])),
            shape=data["adj_shape"],
        )
        y = data["y"]
        self.mask_tr = data["mask_tr"]
        self.mask_va = data["mask_va"]
        self.mask_te = data["mask_te"]

        return [Graph(x=x, a=a, y=y)]

    def download(self):
        print("Downloading {} dataset.".format(self.name))
        url = self.url.format(self.name)
        download_file(url, self.path, self.name + ".zip")

        # Datasets are zipped in a folder: unpack them
        parent = self.path
        subfolder = osp.join(self.path, self.name)
        for filename in os.listdir(subfolder):
            shutil.move(osp.join(subfolder, filename), osp.join(parent, filename))
        os.rmdir(subfolder)

        x, adj, y, mask_tr, mask_va, mask_te = preprocess_data(self.path, self.name)

        # Save pre-processed data
        npz_file = osp.join(self.path, self.name) + ".npz"
        adj = adj.tocoo()
        np.savez(
            npz_file,
            x=x,
            adj_data=adj.data,
            adj_row=adj.row,
            adj_col=adj.col,
            adj_shape=adj.shape,
            y=y,
            mask_tr=mask_tr,
            mask_va=mask_va,
            mask_te=mask_te,
        )

    @staticmethod
    def available_datasets():
        return ["ppi", "reddit"]


class PPI(GraphSage):
    """
    Alias for `GraphSage('ppi')`.
    """

    def __init__(self, **kwargs):
        super().__init__(name="ppi", **kwargs)


class Reddit(GraphSage):
    """
    Alias for `GraphSage('reddit')`.
    """

    def __init__(self, **kwargs):
        super().__init__(name="reddit", **kwargs)


def preprocess_data(path, name):
    """
    Code adapted from https://github.com/williamleif/GraphSAGE
    """
    print("Processing dataset.")
    prefix = osp.join(path, name)

    G_data = json.load(open(prefix + "-G.json"))
    G = json_graph.node_link_graph(G_data)

    x = np.load(prefix + "-feats.npy").astype(np.float32)

    id_map = json.load(open(prefix + "-id_map.json"))
    if list(id_map.keys())[0].isdigit():
        conversion = int
    else:
        conversion = lambda n: n
    id_map = {conversion(k): int(v) for k, v in id_map.items()}
    n = len(id_map)

    class_map = json.load(open(prefix + "-class_map.json"))
    if isinstance(list(class_map.values())[0], list):
        lab_conversion = lambda n: n
    else:
        lab_conversion = int
    class_map = {conversion(k): lab_conversion(v) for k, v in class_map.items()}

    # In Reddit, since edges in G_data are (int, int) and node ids are strings,
    # Networkx weirdly doubles the number of nodes by creating duplicates with integer
    # ids.
    # We create a reverse lookup set to make sure that edges exist when creating the
    # adjacency matrix.
    inverse_ids_lookup = set(id_map.values())

    # Adjacency matrix
    edges = [
        edge
        for edge in G.edges()
        if edge[0] in inverse_ids_lookup and edge[1] in inverse_ids_lookup
    ]
    edges = np.array(edges, dtype=np.int32)
    adj = sp.csr_matrix(
        (np.ones((edges.shape[0]), dtype=np.float32), (edges[:, 0], edges[:, 1])),
        shape=(n, n),
    )
    adj += adj.T
    adj.data = np.clip(adj.data, 0, 1)

    # Process labels
    if isinstance(list(class_map.values())[0], list):
        num_classes = len(list(class_map.values())[0])
        y = np.zeros((n, num_classes), dtype=np.float32)
        for k in class_map.keys():
            y[id_map[k], :] = np.array(class_map[k])
    else:
        num_classes = len(set(class_map.values()))
        y = np.zeros((n, num_classes), dtype=np.float32)
        for k in class_map.keys():
            y[id_map[k], class_map[k]] = 1

    # Get train/val/test indexes
    idx_va = np.array(
        [id_map[k] for k in id_map if k in G.nodes and G.nodes[k]["val"]],
        dtype=np.int32,
    )
    idx_te = np.array(
        [id_map[k] for k in id_map if k in G.nodes and G.nodes[k]["test"]],
        dtype=np.int32,
    )
    mask_tr = np.ones(n, dtype=np.bool)
    mask_va = np.zeros(n, dtype=np.bool)
    mask_te = np.zeros(n, dtype=np.bool)
    mask_tr[idx_va] = False
    mask_tr[idx_te] = False
    mask_va[idx_va] = True
    mask_te[idx_te] = True

    return x, adj, y, mask_tr, mask_va, mask_te



import json
import os
import os.path as osp

import numpy as np
import scipy.sparse as sp

from spektral.data import Dataset, Graph
from spektral.datasets.citation import _preprocess_features
from spektral.datasets.dblp import _download_url
from spektral.utils import label_to_one_hot


class Flickr(Dataset):
    """
    The Flickr dataset from the [Zeng at al. (2019)](https://arxiv.org/abs/1907.04931) paper,
    containing descriptions and common properties of images.

    **Arguments**

    - `normalize_x`: if True, normalize the features.
    - `dtype`: numpy dtype of graph data.
    """

    url = "https://docs.google.com/uc?export=download&id={}&confirm=t"

    adj_full_id = "1crmsTbd1-2sEXsGwa2IKnIB7Zd3TmUsy"
    feats_id = "1join-XdvX3anJU_MLVtick7MgeAQiWIZ"
    class_map_id = "1uxIkbtg5drHTsKt-PAsZZ4_yJmgFmle9"
    role_id = "1htXCtuktuCW8TR8KiKfrFDAxUgekQoV7"

    def __init__(self, normalize_x=False, dtype=np.float32, **kwargs):
        self.dtype = dtype
        self.normalize_x = normalize_x
        super().__init__(**kwargs)

    def download(self):
        print("Downloading Flickr dataset.")
        file_path = _download_url(self.url.format(self.adj_full_id), self.path)
        os.rename(file_path, osp.join(self.path, "adj_full.npz"))

        file_path = _download_url(self.url.format(self.feats_id), self.path)
        os.rename(file_path, osp.join(self.path, "feats.npy"))

        file_path = _download_url(self.url.format(self.class_map_id), self.path)
        os.rename(file_path, osp.join(self.path, "class_map.json"))

        file_path = _download_url(self.url.format(self.role_id), self.path)
        os.rename(file_path, osp.join(self.path, "role.json"))

    def read(self):
        f = np.load(osp.join(self.path, "adj_full.npz"))
        a = sp.csr_matrix((f["data"], f["indices"], f["indptr"]), f["shape"])

        x = np.load(osp.join(self.path, "feats.npy"))

        if self.normalize_x:
            print("Pre-processing node features")
            x = _preprocess_features(x)

        y = np.zeros(x.shape[0])
        with open(osp.join(self.path, "class_map.json")) as f:
            class_map = json.load(f)
            for key, item in class_map.items():
                y[int(key)] = item

        y = label_to_one_hot(y, np.unique(y))

        with open(osp.join(self.path, "role.json")) as f:
            role = json.load(f)

        self.train_mask = np.zeros(x.shape[0], dtype=bool)
        self.train_mask[np.array(role["tr"])] = 1

        self.val_mask = np.zeros(x.shape[0], dtype=bool)
        self.val_mask[np.array(role["va"])] = 1

        self.test_mask = np.zeros(x.shape[0], dtype=bool)
        self.test_mask[np.array(role["te"])] = 1

        return [
            Graph(
                x=x.astype(self.dtype),
                a=a.astype(self.dtype),
                y=y.astype(self.dtype),
            )
        ]



import numpy as np
import scipy.sparse as sp
from scipy.spatial import Delaunay as DelaunaySP


class Delaunay:
    """
    Computes the Delaunay triangulation of the node features.

    The adjacency matrix is obtained from the edges of the triangulation and
    replaces the previous adjacency matrix.
    Duplicate edges are ignored and the adjacency matrix is binary.

    Node features must be 2-dimensional.
    """

    def __call__(self, graph):
        if "x" not in graph:
            raise ValueError("The graph must have node features")
        if graph.n_node_features != 2:
            raise ValueError("Can only compute triangulation for 2-d points.")
        tri = DelaunaySP(graph.x)
        edges = np.concatenate(
            (tri.vertices[:, :2], tri.vertices[:, 1:], tri.vertices[:, ::2]), axis=0
        )
        values = np.ones(edges.shape[0])
        graph.a = sp.csr_matrix((values, edges.T))
        graph.a += graph.a.T
        graph.a.data = np.clip(graph.a.data, 0, 1)

        return graph



import tensorflow as tf


def mixed_mode_support(scatter_fn):
    def _wrapper_mm_support(updates, indices, N):
        if len(updates.shape) == 3:
            updates = tf.transpose(updates, perm=(1, 0, 2))
        out = scatter_fn(updates, indices, N)
        if len(out.shape) == 3:
            out = tf.transpose(out, perm=(1, 0, 2))
        return out

    _wrapper_mm_support.__name__ = scatter_fn.__name__
    return _wrapper_mm_support


@mixed_mode_support
def scatter_sum(messages, indices, n_nodes):
    """
    Sums messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all zeros.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_sum(messages, indices, n_nodes)


@mixed_mode_support
def scatter_mean(messages, indices, n_nodes):
    """
    Averages messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all zeros.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_mean(messages, indices, n_nodes)


@mixed_mode_support
def scatter_max(messages, indices, n_nodes):
    """
    Max-reduces messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be the minimum possible
    value for the dtype of the messages
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_max(messages, indices, n_nodes)


@mixed_mode_support
def scatter_min(messages, indices, n_nodes):
    """
    Min-reduces messages according to the segments defined by `indices`, with
    support for messages in single/disjoint mode (rank 2) and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be the maximum possible
    value for the dtype of the messages.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_min(messages, indices, n_nodes)


@mixed_mode_support
def scatter_prod(messages, indices, n_nodes):
    """
    Multiplies messages element-wise according to the segments defined by
    `indices`, with support for messages in single/disjoint mode (rank 2)
    and mixed mode (rank 3).
    The output has the same rank as the input, with the "nodes" dimension
    changed to `n_nodes`.

    For single/disjoint mode, messages are expected to have shape
    `[n_messages, n_features]` and the output will have shape
    `[n_nodes, n_features]`.

    For mixed mode, messages are expected to have shape
    `[batch, n_messages, n_features]` and the output will have shape
    `[batch, n_nodes, n_features]`.

    Indices are always expected to be a 1D tensor of integers < `n_nodes`, with
    shape '[n_messages]'.
    For any missing index (i.e., any integer `0 <= i < n_nodes` that does not
    appear in `indices`) the corresponding output will be all ones.
    If a given index `i` is negative, it is ignored in the aggregation.

    :param messages: a 2D or 3D Tensor.
    :param indices: A 1D Tensor with indices into the "nodes" dimension of the
    messages.
    :param n_nodes: dimension of the output along the "nodes" dimension.
    :return: a Tensor with the same rank as `messages`.
    """
    return tf.math.unsorted_segment_prod(messages, indices, n_nodes)


OP_DICT = {
    "sum": scatter_sum,
    "mean": scatter_mean,
    "max": scatter_max,
    "min": scatter_min,
    "prod": scatter_prod,
}


def unsorted_segment_softmax(x, indices, n_nodes=None):
    """
    Applies softmax along the segments of a Tensor. This operator is similar
    to the tf.math.segment_* operators, which apply a certain reduction to the
    segments. In this case, the output tensor is not reduced and maintains the
    same shape as the input.
    :param x: a Tensor. The softmax is applied along the first dimension.
    :param indices: a Tensor, indices to the segments.
    :param n_nodes: the number of unique segments in the indices. If `None`,
    n_nodes is calculated as the maximum entry in the indices plus 1.
    :return: a Tensor with the same shape as the input.
    """
    n_nodes = tf.reduce_max(indices) + 1 if n_nodes is None else n_nodes
    e_x = tf.exp(
        x - tf.gather(tf.math.unsorted_segment_max(x, indices, n_nodes), indices)
    )
    e_x /= tf.gather(
        tf.math.unsorted_segment_sum(e_x, indices, n_nodes) + 1e-9, indices
    )
    return e_x


def serialize_scatter(identifier):
    if identifier in OP_DICT:
        return identifier
    elif hasattr(identifier, "__name__"):
        for k, v in OP_DICT.items():
            if v.__name__ == identifier.__name__:
                return k
        return None


def deserialize_scatter(scatter):
    if isinstance(scatter, str) and scatter in OP_DICT:
        return OP_DICT[scatter]
    elif callable(scatter):
        return scatter
    else:
        raise ValueError(f"scatter must be callable or string: {list(OP_DICT.keys())}")



from tensorflow.keras import activations
from tensorflow.keras.layers import Dense, Dropout
from tensorflow.keras.models import Sequential

from spektral.layers import ops
from spektral.layers.convolutional.conv import Conv
from spektral.utils import gcn_filter


class APPNPConv(Conv):
    r"""
    The APPNP operator from the paper

    > [Predict then Propagate: Graph Neural Networks meet Personalized PageRank](https://arxiv.org/abs/1810.05997)<br>
    > Johannes Klicpera et al.

    **Mode**: single, disjoint, mixed, batch.

    This layer computes:
    $$
        \Z^{(0)} = \textrm{MLP}(\X); \\
        \Z^{(K)} = (1 - \alpha) \hat \D^{-1/2} \hat \A \hat \D^{-1/2} \Z^{(K - 1)} +
                   \alpha \Z^{(0)},
    $$
    where \(\alpha\) is the teleport probability, \(\textrm{MLP}\) is a
    multi-layer perceptron, and \(K\) is defined by the `propagations` argument.

    **Input**

    - Node features of shape `([batch], n_nodes, n_node_features)`;
    - Modified Laplacian of shape `([batch], n_nodes, n_nodes)`; can be computed with
    `spektral.utils.convolution.gcn_filter`.

    **Output**

    - Node features with the same shape as the input, but with the last
    dimension changed to `channels`.

    **Arguments**

    - `channels`: number of output channels;
    - `alpha`: teleport probability during propagation;
    - `propagations`: number of propagation steps;
    - `mlp_hidden`: list of integers, number of hidden units for each hidden
    layer in the MLP (if None, the MLP has only the output layer);
    - `mlp_activation`: activation for the MLP layers;
    - `dropout_rate`: dropout rate for Laplacian and MLP layers;
    - `activation`: activation function;
    - `use_bias`: bool, add a bias vector to the output;
    - `kernel_initializer`: initializer for the weights;
    - `bias_initializer`: initializer for the bias vector;
    - `kernel_regularizer`: regularization applied to the weights;
    - `bias_regularizer`: regularization applied to the bias vector;
    - `activity_regularizer`: regularization applied to the output;
    - `kernel_constraint`: constraint applied to the weights;
    - `bias_constraint`: constraint applied to the bias vector.
    """

    def __init__(
        self,
        channels,
        alpha=0.2,
        propagations=1,
        mlp_hidden=None,
        mlp_activation="relu",
        dropout_rate=0.0,
        activation=None,
        use_bias=True,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        activity_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            activation=activation,
            use_bias=use_bias,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            activity_regularizer=activity_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )
        self.channels = channels
        self.mlp_hidden = mlp_hidden if mlp_hidden else []
        self.alpha = alpha
        self.propagations = propagations
        self.mlp_activation = activations.get(mlp_activation)
        self.dropout_rate = dropout_rate

    def build(self, input_shape):
        assert len(input_shape) >= 2
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
            dtype=self.dtype,
        )
        mlp_layers = []
        for channels in self.mlp_hidden:
            mlp_layers.extend(
                [
                    Dropout(self.dropout_rate),
                    Dense(channels, self.mlp_activation, **layer_kwargs),
                ]
            )
        mlp_layers.append(Dense(self.channels, "linear", **layer_kwargs))
        self.mlp = Sequential(mlp_layers)
        self.built = True

    def call(self, inputs, mask=None):
        x, a = inputs

        mlp_out = self.mlp(x)
        output = mlp_out
        for _ in range(self.propagations):
            output = (1 - self.alpha) * ops.modal_dot(a, output) + self.alpha * mlp_out

        if mask is not None:
            output *= mask[0]
        output = self.activation(output)

        return output

    @property
    def config(self):
        return {
            "channels": self.channels,
            "alpha": self.alpha,
            "propagations": self.propagations,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": activations.serialize(self.mlp_activation),
            "dropout_rate": self.dropout_rate,
        }

    @staticmethod
    def preprocess(a):
        return gcn_filter(a)



import tensorflow as tf
from tensorflow.keras import Sequential
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Dense

from spektral.layers import ops
from spektral.layers.pooling.src import SRCPool


class MinCutPool(SRCPool):
    r"""
    A MinCut pooling layer from the paper

    > [Spectral Clustering with Graph Neural Networks for Graph Pooling](https://arxiv.org/abs/1907.00481)<br>
    > Filippo Maria Bianchi et al.

    **Mode**: single, batch.

    This layer learns a soft clustering of the input graph as follows:
    $$
    \begin{align}
        \S &= \textrm{MLP}(\X); \\
        \X' &= \S^\top \X \\
        \A' &= \S^\top \A \S; \\
    \end{align}
    $$
    where \(\textrm{MLP}\) is a multi-layer perceptron with softmax output.

    Two auxiliary loss terms are also added to the model: the minimum cut loss
    $$
        L_c = - \frac{ \mathrm{Tr}(\S^\top \A \S) }{ \mathrm{Tr}(\S^\top \D \S) }
    $$
    and the orthogonality loss
    $$
        L_o = \left\|
            \frac{\S^\top \S}{\| \S^\top \S \|_F}
            - \frac{\I_K}{\sqrt{K}}
        \right\|_F.
    $$

    The layer can be used without a supervised loss to compute node clustering by
    minimizing the two auxiliary losses.

    **Input**

    - Node features of shape `(batch, n_nodes_in, n_node_features)`;
    - Symmetrically normalized adjacency matrix of shape
    `(batch, n_nodes_in, n_nodes_in)`;

    **Output**

    - Reduced node features of shape `(batch, n_nodes_out, n_node_features)`;
    - Reduced adjacency matrix of shape `(batch, n_nodes_out, n_nodes_out)`;
    - If `return_selection=True`, the selection matrix of shape
    `(batch, n_nodes_in, n_nodes_out)`.

    **Arguments**

    - `k`: number of output nodes;
    - `mlp_hidden`: list of integers, number of hidden units for each hidden layer in
    the MLP used to compute cluster assignments (if `None`, the MLP has only one output
    layer);
    - `mlp_activation`: activation for the MLP layers;
    - `return_selection`: boolean, whether to return the selection matrix;
    - `use_bias`: use bias in the MLP;
    - `kernel_initializer`: initializer for the weights of the MLP;
    - `bias_initializer`: initializer for the bias of the MLP;
    - `kernel_regularizer`: regularization applied to the weights of the MLP;
    - `bias_regularizer`: regularization applied to the bias of the MLP;
    - `kernel_constraint`: constraint applied to the weights of the MLP;
    - `bias_constraint`: constraint applied to the bias of the MLP;
    """

    def __init__(
        self,
        k,
        mlp_hidden=None,
        mlp_activation="relu",
        return_selection=False,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            return_selection=return_selection,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )

        self.k = k
        self.mlp_hidden = mlp_hidden if mlp_hidden is not None else []
        self.mlp_activation = mlp_activation

    def build(self, input_shape):
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )
        self.mlp = Sequential(
            [
                Dense(channels, self.mlp_activation, **layer_kwargs)
                for channels in self.mlp_hidden
            ]
            + [Dense(self.k, "softmax", **layer_kwargs)]
        )

        super().build(input_shape)

    def call(self, inputs, mask=None):
        x, a, i = self.get_inputs(inputs)
        return self.pool(x, a, i, mask=mask)

    def select(self, x, a, i, mask=None):
        s = self.mlp(x)
        if mask is not None:
            s *= mask[0]

        # Orthogonality loss
        ortho_loss = self.orthogonality_loss(s)
        if K.ndim(a) == 3:
            ortho_loss = K.mean(ortho_loss)
        self.add_loss(ortho_loss)

        return s

    def reduce(self, x, s, **kwargs):
        return ops.modal_dot(s, x, transpose_a=True)

    def connect(self, a, s, **kwargs):
        a_pool = ops.matmul_at_b_a(s, a)

        # MinCut loss
        cut_loss = self.mincut_loss(a, s, a_pool)
        if K.ndim(a) == 3:
            cut_loss = K.mean(cut_loss)
        self.add_loss(cut_loss)

        # Post-processing of A
        a_pool = tf.linalg.set_diag(
            a_pool, tf.zeros(K.shape(a_pool)[:-1], dtype=a_pool.dtype)
        )
        a_pool = ops.normalize_A(a_pool)

        return a_pool

    def reduce_index(self, i, s, **kwargs):
        i_mean = tf.math.segment_mean(i, i)
        i_pool = ops.repeat(i_mean, tf.ones_like(i_mean) * self.k)

        return i_pool

    def orthogonality_loss(self, s):
        ss = ops.modal_dot(s, s, transpose_a=True)
        i_s = tf.eye(self.k, dtype=ss.dtype)
        ortho_loss = tf.norm(
            ss / tf.norm(ss, axis=(-1, -2), keepdims=True) - i_s / tf.norm(i_s),
            axis=(-1, -2),
        )
        return ortho_loss

    @staticmethod
    def mincut_loss(a, s, a_pool):
        num = tf.linalg.trace(a_pool)
        d = ops.degree_matrix(a)
        den = tf.linalg.trace(ops.matmul_at_b_a(s, d))
        cut_loss = -(num / den)
        return cut_loss

    def get_config(self):
        config = {
            "k": self.k,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": self.mlp_activation,
        }
        base_config = super().get_config()
        return {**base_config, **config}



import numpy as np

from spektral.utils import one_hot


class Degree:
    """
    Concatenates to each node attribute the one-hot degree of the corresponding
    node.

    The adjacency matrix is expected to have integer entries and the degree is
    cast to integer before one-hot encoding.

    **Arguments**

    - `max_degree`: the maximum degree of the nodes, i.e., the size of the
    one-hot vectors.
    """

    def __init__(self, max_degree):
        self.max_degree = max_degree

    def __call__(self, graph):
        if "a" not in graph:
            raise ValueError("The graph must have an adjacency matrix")
        degree = graph.a.sum(1).astype(int)
        if isinstance(degree, np.matrix):
            degree = np.asarray(degree)[:, 0]
        degree = one_hot(degree, self.max_degree + 1)
        if "x" not in graph:
            graph.x = degree
        else:
            graph.x = np.concatenate((graph.x, degree), axis=-1)

        return graph


class MaxDegree:
    def __call__(self, graph):
        return graph.a.sum(1).max()



from .diff_pool import DiffPool
from .dmon_pool import DMoNPool
from .global_pool import (
    GlobalAttentionPool,
    GlobalAttnSumPool,
    GlobalAvgPool,
    GlobalMaxPool,
    GlobalSumPool,
    SortPool,
)
from .just_balance_pool import JustBalancePool
from .la_pool import LaPool
from .mincut_pool import MinCutPool
from .sag_pool import SAGPool
from .src import SRCPool
from .topk_pool import TopKPool



import numpy as np
import scipy.sparse as sp
import tensorflow as tf

from spektral import layers
from spektral.utils.sparse import sp_matrix_to_sp_tensor
from tests.test_layers.convolutional.core import _test_get_config

tol = 1e-6


def test_disjoint_2_batch():
    X = np.array([[1, 0], [0, 1], [1, 1], [0, 0], [1, 2]])
    I = np.array([0, 0, 0, 1, 1])
    A_data = [1, 1, 1, 1, 1]
    A_row = [0, 1, 2, 3, 4]
    A_col = [1, 0, 1, 4, 3]
    A = sp_matrix_to_sp_tensor(sp.csr_matrix((A_data, (A_row, A_col)), shape=(5, 5)))

    expected_X = np.array(
        [[[1.0, 0.0], [0.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [1.0, 2.0], [0.0, 0.0]]]
    )
    expected_A = np.array(
        [
            [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 1.0, 0.0]],
            [[0.0, 1.0, 0.0], [1.0, 0.0, 0.0], [0.0, 0.0, 0.0]],
        ]
    )

    result_X, result_A = layers.Disjoint2Batch()((X, A, I))
    assert np.allclose(result_A, expected_A, atol=tol)
    assert np.allclose(result_X, expected_X, atol=tol)

    _test_get_config(layers.Disjoint2Batch)


def test_sparse_dropout():
    """Ensure SparseDropout.sparse_dropout gradients don't throw issues."""
    n = 5
    rate = 0.5
    with tf.GradientTape() as tape:
        values = tf.Variable(tf.random.uniform((n,)))
        tape.watch(values)
        st = tf.sparse.eye(5)
        st = tf.SparseTensor(st.indices, values, st.dense_shape)
        st = layers.SparseDropout(rate)(st, training=True)
        loss = tf.sparse.reduce_sum(st)
    grad = tape.gradient(loss, values)
    assert grad is not None

    _test_get_config(layers.SparseDropout, rate=rate)


def test_inner_products():
    x = np.random.rand(10, 2)

    layer = layers.InnerProduct(trainable_kernel=True)
    output = layer(x)
    assert output.shape == (10, 10)

    layer = layers.InnerProduct(trainable_kernel=False)
    output = layer(x)
    assert output.shape == (10, 10)

    _test_get_config(layers.InnerProduct, trainable_kernel=True)

    layer = layers.MinkowskiProduct()
    output = layer(x)
    assert output.shape == (10, 10)

    _test_get_config(layers.MinkowskiProduct)



import tensorflow as tf
from tensorflow.keras import Sequential
from tensorflow.keras import backend as K
from tensorflow.keras.layers import Dense

from spektral.layers import ops
from spektral.layers.pooling.src import SRCPool


class DMoNPool(SRCPool):
    r"""
    The DMoN pooling layer from the paper

    > [Graph Clustering with Graph Neural Networks](https://arxiv.org/abs/2006.16904)<br>
    > Anton Tsitsulin et al.

    **Mode**: single, batch.

    This layer learns a soft clustering of the input graph as follows:
    $$
    \begin{align}
        \C &= \textrm{MLP}(\X); \\
        \X' &= \C^\top \X \\
        \A' &= \C^\top \A \C; \\
    \end{align}
    $$
    where \(\textrm{MLP}\) is a multi-layer perceptron with softmax output.

    Two auxiliary loss terms are also added to the model: the modularity loss
    $$
        L_m = - \frac{1}{2m} \mathrm{Tr}(\C^\top \A \C - \C^\top \d^\top \d \C)
    $$
    and the collapse regularization loss
    $$
        L_c = \frac{\sqrt{k}}{n} \left\|
            \sum_i \C_i^\top
        \right\|_F -1.
    $$

    This layer is based on the original implementation found
    [here](https://github.com/google-research/google-research/blob/master/graph_embedding/dmon/dmon.py).

    **Input**

    - Node features of shape `(batch, n_nodes_in, n_node_features)`;
    - Symmetrically normalized adjacency matrix of shape
    `(batch, n_nodes_in, n_nodes_in)`;

    **Output**

    - Reduced node features of shape `(batch, n_nodes_out, n_node_features)`;
    - Reduced adjacency matrix of shape `(batch, n_nodes_out, n_nodes_out)`;
    - If `return_selection=True`, the selection matrix of shape
    `(batch, n_nodes_in, n_nodes_out)`.

    **Arguments**

    - `k`: number of output nodes;
    - `mlp_hidden`: list of integers, number of hidden units for each hidden layer in
    the MLP used to compute cluster assignments (if `None`, the MLP has only one output
    layer);
    - `mlp_activation`: activation for the MLP layers;
    - `collapse_regularization`: strength of the collapse regularization;
    - `return_selection`: boolean, whether to return the selection matrix;
    - `use_bias`: use bias in the MLP;
    - `kernel_initializer`: initializer for the weights of the MLP;
    - `bias_initializer`: initializer for the bias of the MLP;
    - `kernel_regularizer`: regularization applied to the weights of the MLP;
    - `bias_regularizer`: regularization applied to the bias of the MLP;
    - `kernel_constraint`: constraint applied to the weights of the MLP;
    - `bias_constraint`: constraint applied to the bias of the MLP;
    """

    def __init__(
        self,
        k,
        mlp_hidden=None,
        mlp_activation="relu",
        return_selection=False,
        collapse_regularization=0.1,
        kernel_initializer="glorot_uniform",
        bias_initializer="zeros",
        kernel_regularizer=None,
        bias_regularizer=None,
        kernel_constraint=None,
        bias_constraint=None,
        **kwargs
    ):
        super().__init__(
            k=k,
            mlp_hidden=mlp_hidden,
            mlp_activation=mlp_activation,
            return_selection=return_selection,
            kernel_initializer=kernel_initializer,
            bias_initializer=bias_initializer,
            kernel_regularizer=kernel_regularizer,
            bias_regularizer=bias_regularizer,
            kernel_constraint=kernel_constraint,
            bias_constraint=bias_constraint,
            **kwargs
        )

        self.k = k
        self.mlp_hidden = mlp_hidden if mlp_hidden is not None else []
        self.mlp_activation = mlp_activation
        self.collapse_regularization = collapse_regularization

    def build(self, input_shape):
        layer_kwargs = dict(
            kernel_initializer=self.kernel_initializer,
            bias_initializer=self.bias_initializer,
            kernel_regularizer=self.kernel_regularizer,
            bias_regularizer=self.bias_regularizer,
            kernel_constraint=self.kernel_constraint,
            bias_constraint=self.bias_constraint,
        )
        self.mlp = Sequential(
            [
                Dense(channels, self.mlp_activation, **layer_kwargs)
                for channels in self.mlp_hidden
            ]
            + [Dense(self.k, "softmax", **layer_kwargs)]
        )

        super().build(input_shape)

    def call(self, inputs, mask=None):
        x, a, i = self.get_inputs(inputs)
        return self.pool(x, a, i, mask=mask)

    def select(self, x, a, i, mask=None):
        s = self.mlp(x)
        if mask is not None:
            s *= mask[0]

        # Collapse loss
        col_loss = self.collapse_loss(a, s)
        if K.ndim(a) == 3:
            col_loss = K.mean(col_loss)
        self.add_loss(self.collapse_regularization * col_loss)

        return s

    def reduce(self, x, s, **kwargs):
        return ops.modal_dot(s, x, transpose_a=True)

    def connect(self, a, s, **kwargs):
        a_pool = ops.matmul_at_b_a(s, a)

        # Modularity loss
        mod_loss = self.modularity_loss(a, s, a_pool)
        if K.ndim(a) == 3:
            mod_loss = K.mean(mod_loss)
        self.add_loss(mod_loss)

        return a_pool

    def reduce_index(self, i, s, **kwargs):
        i_mean = tf.math.segment_mean(i, i)
        i_pool = ops.repeat(i_mean, tf.ones_like(i_mean) * self.k)

        return i_pool

    def modularity_loss(self, a, s, a_pool):

        if K.is_sparse(a):
            n_edges = tf.cast(len(a.values), dtype=s.dtype)

            degrees = tf.sparse.reduce_sum(a, axis=-1)
            degrees = tf.reshape(degrees, (-1, 1))
        else:
            n_edges = tf.cast(tf.math.count_nonzero(a, axis=(-2, -1)), dtype=s.dtype)
            degrees = tf.reduce_sum(a, axis=-1, keepdims=True)

        normalizer_left = tf.matmul(s, degrees, transpose_a=True)
        normalizer_right = tf.matmul(degrees, s, transpose_a=True)

        if K.ndim(s) == 3:
            normalizer = (
                ops.modal_dot(normalizer_left, normalizer_right)
                / 2
                / tf.reshape(n_edges, [tf.shape(n_edges)[0]] + [1] * 2)
            )
        else:
            normalizer = ops.modal_dot(normalizer_left, normalizer_right) / 2 / n_edges

        loss = -tf.linalg.trace(a_pool - normalizer) / 2 / n_edges

        return loss

    def collapse_loss(self, a, s):
        cluster_sizes = tf.math.reduce_sum(s, axis=-2)
        n_nodes = tf.cast(tf.shape(a)[-1], s.dtype)
        loss = (
            tf.norm(cluster_sizes, axis=-1)
            / n_nodes
            * tf.sqrt(tf.cast(self.k, s.dtype))
            - 1
        )

        return loss

    def get_config(self):
        config = {
            "collapse_regularization": self.collapse_regularization,
            "k": self.k,
            "mlp_hidden": self.mlp_hidden,
            "mlp_activation": self.mlp_activation,
        }
        base_config = super().get_config()
        return {**base_config, **config}



import copy
import os.path as osp
import warnings

import numpy as np
import tensorflow as tf

from spektral.data.graph import Graph
from spektral.data.utils import get_spec
from spektral.datasets.utils import DATASET_FOLDER


class Dataset:
    """
    A container for Graph objects. This class can be extended to represent a
    graph dataset.

    To create a `Dataset`, you must implement the `Dataset.read()` method, which
    must return a list of `spektral.data.Graph` objects:

    ```py
    class MyDataset(Dataset):
        def read(self):
            return [Graph(x=x, adj=adj, y=y) for x, adj, y in some_magic_list]
    ```

    The `download()` method is automatically called if the path returned by
    `Dataset.path` does not exists (default `~/.spektral/datasets/ClassName/`).

    In this case, `download()` will be called before `read()`.

    Datasets should generally behave like Numpy arrays for any operation that
    uses simple 1D indexing:

    ```py
    >>> dataset[0]
    Graph(...)

    >>> dataset[[1, 2, 3]]
    Dataset(n_graphs=3)

    >>> dataset[1:10]
    Dataset(n_graphs=9)

    >>> np.random.shuffle(dataset)  # shuffle in-place

    >>> for graph in dataset[:3]:
    >>>     print(graph)
    Graph(...)
    Graph(...)
    Graph(...)
    ```

    Datasets have the following properties that are automatically computed:

        - `n_nodes`: the number of nodes in the dataset (always None, except
        in single and mixed mode datasets);
        - `n_node_features`: the size of the node features (assumed to be equal
        for all graphs);
        - `n_edge_features`: the size of the edge features (assumed to be equal
        for all graphs);
        - `n_labels`: the size of the labels (assumed to be equal for all
        graphs); this is computed as `y.shape[-1]`.

    Any additional `kwargs` passed to the constructor will be automatically
    assigned as instance attributes of the dataset.

    Datasets also offer three main manipulation functions to apply callables to
    their graphs:

    - `apply(transform)`: replaces each graph with the output of `transform(graph)`.
    See `spektral.transforms` for some ready-to-use transforms.<br>
    Example: `apply(spektral.transforms.NormalizeAdj())` normalizes the
    adjacency matrix of each graph in the dataset.
    - `map(transform, reduce=None)`: returns a list containing the output
    of `transform(graph)` for each graph. If `reduce` is a `callable`, then
    returns `reduce(output_list)`.<br>
    Example: `map(lambda: g.n_nodes, reduce=np.mean)` will return the
    average number of nodes in the dataset.
    - `filter(function)`: removes from the dataset any graph for which
    `function(graph) is False`.<br>
    Example: `filter(lambda: g.n_nodes < 100)` removes from the dataset all
    graphs bigger than 100 nodes.

    Datasets in mixed mode (one adjacency matrix, many instances of node features)
    are expected to have a particular structure.
    The graphs returned by `read()` should not have an adjacency matrix,
    which should be instead stored as a singleton in the dataset's `a` attribute.
    For example:

    ```py
    class MyMixedModeDataset(Dataset):
        def read(self):
            self.a = compute_adjacency_matrix()
            return [Graph(x=x, y=y) for x, y in some_magic_list]
    ```

    Have a look at the `spektral.datasets` module for examples of popular
    datasets already implemented.

    **Arguments**

    - `transforms`: a callable or list of callables that are automatically
    applied to the graphs after loading the dataset.
    """

    def __init__(self, transforms=None, **kwargs):
        self.a = None  # Used for mixed-mode datasets
        # Read extra kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

        # Download data
        if not osp.exists(self.path):
            self.download()

        # Read graphs
        self.graphs = self.read()
        if self.a is None and self.__len__() > 0 and "a" not in self.graphs[0]:
            warnings.warn(
                "The graphs in this dataset have no adjacency matrix. "
                "Is this intentional?"
            )

        # Apply transforms
        if transforms is not None:
            if not isinstance(transforms, (list, tuple)) and callable(transforms):
                transforms = [transforms]
            elif not all([callable(t) for t in transforms]):
                raise ValueError(
                    "`transforms` must be a callable or list of " "callables"
                )
            else:
                pass
            for t in transforms:
                self.apply(t)

    def read(self):
        raise NotImplementedError

    def download(self):
        pass

    def apply(self, transform):
        if not callable(transform):
            raise ValueError("`transform` must be callable")

        for i in range(len(self.graphs)):
            self.graphs[i] = transform(self.graphs[i])

    def map(self, transform, reduce=None):
        if not callable(transform):
            raise ValueError("`transform` must be callable")
        if reduce is not None and not callable(reduce):
            raise ValueError("`reduce` must be callable")

        out = [transform(g) for g in self.graphs]
        return reduce(out) if reduce is not None else out

    def filter(self, function):
        if not callable(function):
            raise ValueError("`function` must be callable")
        self.graphs = [g for g in self.graphs if function(g)]

    def __getitem__(self, key):
        if not (
            np.issubdtype(type(key), np.integer)
            or isinstance(key, (slice, list, tuple, np.ndarray))
        ):
            raise ValueError("Unsupported key type: {}".format(type(key)))
        if np.issubdtype(type(key), np.integer):
            return self.graphs[int(key)]
        else:
            dataset = copy.copy(self)
            if isinstance(key, slice):
                dataset.graphs = self.graphs[key]
            else:
                dataset.graphs = [self.graphs[i] for i in key]
            return dataset

    def __setitem__(self, key, value):
        is_iterable = isinstance(value, (list, tuple))
        if not isinstance(value, (Graph, list, tuple)):
            raise ValueError(
                "Datasets can only be assigned Graphs or " "sequences of Graphs"
            )
        if is_iterable and not all([isinstance(v, Graph) for v in value]):
            raise ValueError("Assigned sequence must contain only Graphs")
        if is_iterable and isinstance(key, int):
            raise ValueError("Cannot assign multiple Graphs to one location")
        if not is_iterable and isinstance(key, (slice, list, tuple)):
            raise ValueError("Cannot assign one Graph to multiple locations")
        if not (isinstance(key, (int, slice, list, tuple))):
            raise ValueError("Unsupported key type: {}".format(type(key)))

        if isinstance(key, int):
            self.graphs[key] = value
        else:
            if isinstance(key, slice):
                self.graphs[key] = value
            else:
                for i, k in enumerate(key):
                    self.graphs[k] = value[i]

    def __add__(self, other):
        self_new = copy.copy(self)
        self_new.graphs = self.graphs + other.graphs

        return self_new

    def __len__(self):
        return len(self.graphs)

    def __repr__(self):
        return "{}(n_graphs={})".format(self.__class__.__name__, self.n_graphs)

    @property
    def path(self):
        return osp.join(DATASET_FOLDER, self.__class__.__name__)

    @property
    def n_graphs(self):
        return self.__len__()

    @property
    def n_nodes(self):
        if len(self.graphs) == 1:
            return self.graphs[0].n_nodes
        elif self.a is not None:
            return self.a.shape[-1]
        else:
            return None

    @property
    def n_node_features(self):
        if len(self.graphs) >= 1:
            return self.graphs[0].n_node_features
        else:
            return None

    @property
    def n_edge_features(self):
        if len(self.graphs) >= 1:
            return self.graphs[0].n_edge_features
        else:
            return None

    @property
    def n_labels(self):
        if len(self.graphs) >= 1:
            return self.graphs[0].n_labels
        else:
            return None

    @property
    def signature(self):
        """
        This property computes the signature of the dataset, which can be
        passed to `spektral.data.utils.to_tf_signature(signature)` to compute
        the TensorFlow signature. You can safely ignore this property unless
        you are creating a custom `Loader`.

        A signature consist of the TensorFlow TypeSpec, shape, and dtype of
        all characteristic matrices of the graphs in the Dataset. This is
        returned as a dictionary of dictionaries, with keys `x`, `a`, `e`, and
        `y` for the four main data matrices.

        Each sub-dictionary will have keys `spec`, `shape` and `dtype`.
        """
        if len(self.graphs) == 0:
            return None
        signature = {}
        graph = self.graphs[0]  # This is always non-empty
        if graph.x is not None:
            signature["x"] = dict()
            signature["x"]["spec"] = get_spec(graph.x)
            signature["x"]["shape"] = (None, self.n_node_features)
            signature["x"]["dtype"] = tf.as_dtype(graph.x.dtype)
        if graph.a is not None:
            signature["a"] = dict()
            signature["a"]["spec"] = get_spec(graph.a)
            signature["a"]["shape"] = (None, None)
            signature["a"]["dtype"] = tf.as_dtype(graph.a.dtype)
        if graph.e is not None:
            signature["e"] = dict()
            signature["e"]["spec"] = get_spec(graph.e)
            signature["e"]["shape"] = (None, self.n_edge_features)
            signature["e"]["dtype"] = tf.as_dtype(graph.e.dtype)
        if graph.y is not None:
            signature["y"] = dict()
            signature["y"]["spec"] = get_spec(graph.y)
            signature["y"]["shape"] = (self.n_labels,)
            signature["y"]["dtype"] = tf.as_dtype(np.array(graph.y).dtype)
        return signature



import ast
import sys

import joblib
import networkx as nx
import numpy as np
import pandas as pd
import scipy.sparse as sp

from spektral.data.graph import Graph


def load_binary(filename):
    """
    Loads a pickled file.
    :param filename: a string or file-like object
    :return: the loaded object
    """
    try:
        return joblib.load(filename)
    except ValueError:
        import pickle

        with open(filename, "rb") as f:
            return pickle.load(f, encoding="latin1")


def dump_binary(obj, filename):
    """
    Pickles and saves an object to file.
    :param obj: the object to save
    :param filename: a string or file-like object
    """
    joblib.dump(obj, filename)


def load_csv(filename, **kwargs):
    """
    Loads a csv file with pandas.
    :param filename: a string or file-like object
    :return: the loaded csv
    """
    return pd.read_csv(filename, **kwargs)


def dump_csv(df, filename, convert=False, **kwargs):
    """
    Dumps a pd.DataFrame to csv.
    :param df: the pd.DataFrame to save or equivalent object
    :param filename: a string or file-like object
    :param convert: whether to attempt to convert the given object to
    pd.DataFrame before saving the csv.
    """
    if convert:
        df = pd.DataFrame(df)
    assert hasattr(df, "to_csv"), (
        "Trying to dump object of class {} to csv while pd.DataFrame is "
        "expected. To attempt automatic conversion, set "
        "convert=True.".format(df.__class__)
    )
    df.to_csv(filename, **kwargs)


def load_dot(filename, force_graph=True):
    """
    Loads a graph saved in .dot format.
    :param filename: a string or file-like object
    :param force_graph: whether to force a conversion to nx.Graph after loading.
    This may be useful in the case of .dot files being loaded as nx.MultiGraph.
    :return: the loaded graph
    """
    output = nx.nx_agraph.read_dot(filename)
    if force_graph:
        output = nx.Graph(output)

    for elem in output.nodes().values():
        for k, v in elem.items():
            try:
                elem[k] = ast.literal_eval(v)
            except ValueError:
                elem[k] = str(v)
            except SyntaxError:
                # Probably a numpy array
                elem[k] = np.array(
                    " ".join(v.lstrip("[").rstrip("]").split()).split(" ")
                ).astype(np.float)

    for elem in output.edges().values():
        for k, v in elem.items():
            try:
                elem[k] = ast.literal_eval(v)
            except ValueError:
                elem[k] = str(v)

    return output


def dump_dot(obj, filename):
    """
    Dumps a nx.Graph to .dot file
    :param obj: the nx.Graph (or equivalent) to save
    :param filename: a string or file-like object
    """
    nx.nx_agraph.write_dot(obj, filename)


def load_npy(filename):
    """
    Loads a file saved by np.save.
    :param filename: a string or file-like object
    :return: the loaded object
    """
    if sys.version_info[0] == 3:
        return np.load(filename, encoding="latin1")
    else:
        return np.load(filename)


def dump_npy(obj, filename, zipped=False):
    """
    Saves an object to file using the numpy format.
    :param obj: the object to save
    :param filename: a string or file-like object
    :param zipped: boolean, whether to save the object in the zipped format .npz
    rather than .npy
    """
    if zipped:
        np.savez(filename, obj)
    else:
        np.save(filename, obj)


def load_txt(filename, **kwargs):
    """
    Loads a txt file using np.loadtxt.
    :param filename: a string or file-like object
    :return: the loaded object
    """
    return np.loadtxt(filename, **kwargs)


def dump_txt(obj, filename, **kwargs):
    """
    Saves an object to text file using np.savetxt.
    :param obj: the object to save
    :param filename: a string or file-like object
    """
    np.savetxt(filename, obj, **kwargs)


def _parse_off(lines):
    n_verts, n_faces, _ = map(int, lines[0].split(" "))

    # Read vertices
    verts = np.array(
        [line.strip().split(" ") for line in lines[1 : n_verts + 1]]
    ).astype(float)

    # Read faces
    faces = lines[n_verts + 1 : n_verts + 1 + n_faces]
    faces = [list(map(int, f.split(" "))) for f in faces]
    triangles = np.array(list(filter(lambda f: len(f) == 4, faces))).astype(int)
    rectangles = np.array(list(filter(lambda f: len(f) == 5, faces))).astype(int)
    if len(rectangles) > 0:
        tri_a = rectangles[:, [1, 2, 3]]
        tri_b = rectangles[:, [1, 2, 4]]
        triangles = np.vstack((triangles, tri_a, tri_b))
    triangles = triangles[:, 1:]
    triangles = triangles[triangles[:, 0].argsort()]

    return verts, triangles


def load_off(filename):
    """
    Reads an .off file into a Graph object. Node attributes are the 3d
    coordinates of the points, all faces are converted to edges.
    :param filename: path to the .off file.
    :return: a Graph
    """
    lines = open(filename, "r").read().lstrip("OF\n").splitlines()
    x, faces = _parse_off(lines)
    n = x.shape[0]
    row, col = np.vstack((faces[:, :2], faces[:, 1:], faces[:, ::2])).T
    adj = sp.csr_matrix((np.ones_like(row), (row, col)), shape=(n, n)).tolil()
    adj[col, row] = adj[row, col]
    adj = adj.T.tocsr()
    adj.data = np.clip(adj.data, 0, 1)

    return Graph(x=x, a=adj)


# Reference for implementation:
# # http://www.nonlinear.com/progenesis/sdf-studio/v0.9/faq/sdf-file-format-guidance.aspx
#
# While parsing the SDF file, molecules are stored in a dictionary like this:
#
# {'atoms': [{'atomic_num': 7,
#             'charge': 0,
#             'coords': array([-0.0299,  1.2183,  0.2994]),
#             'index': 0,
#             'info': array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
#             'iso': 0},
#            ...,
#            {'atomic_num': 1,
#             'charge': 0,
#             'coords': array([ 0.6896, -2.3002, -0.1042]),
#             'index': 14,
#             'info': array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0]),
#             'iso': 0}],
#  'bonds': [{'end_atom': 13,
#             'info': array([0, 0, 0]),
#             'start_atom': 4,
#             'stereo': 0,
#             'type': 1},
#            ...,
#            {'end_atom': 8,
#             'info': array([0, 0, 0]),
#             'start_atom': 7,
#             'stereo': 0,
#             'type': 3}],
#  'comment': '',
#  'data': [''],
#  'details': '-OEChem-03231823253D',
#  'n_atoms': 15,
#  'n_bonds': 15,
#  'name': 'gdb_54964',
#  'properties': []}
HEADER_SIZE = 3
NUM_TO_SYMBOL = {
    1: "H",
    2: "He",
    3: "Li",
    4: "Be",
    5: "B",
    6: "C",
    7: "N",
    8: "O",
    9: "F",
    10: "Ne",
    11: "Na",
    12: "Mg",
    13: "Al",
    14: "Si",
    15: "P",
    16: "S",
    17: "Cl",
    18: "Ar",
    19: "K",
    20: "Ca",
    21: "Sc",
    22: "Ti",
    23: "V",
    24: "Cr",
    25: "Mn",
    26: "Fe",
    27: "Co",
    28: "Ni",
    29: "Cu",
    30: "Zn",
    31: "Ga",
    32: "Ge",
    33: "As",
    34: "Se",
    35: "Br",
    36: "Kr",
    37: "Rb",
    38: "Sr",
    39: "Y",
    40: "Zr",
    41: "Nb",
    42: "Mo",
    43: "Tc",
    44: "Ru",
    45: "Rh",
    46: "Pd",
    47: "Ag",
    48: "Cd",
    49: "In",
    50: "Sn",
    51: "Sb",
    52: "Te",
    53: "I",
    54: "Xe",
    55: "Cs",
    56: "Ba",
    57: "La",
    58: "Ce",
    59: "Pr",
    60: "Nd",
    61: "Pm",
    62: "Sm",
    63: "Eu",
    64: "Gd",
    65: "Tb",
    66: "Dy",
    67: "Ho",
    68: "Er",
    69: "Tm",
    70: "Yb",
    71: "Lu",
    72: "Hf",
    73: "Ta",
    74: "W",
    75: "Re",
    76: "Os",
    77: "Ir",
    78: "Pt",
    79: "Au",
    80: "Hg",
    81: "Tl",
    82: "Pb",
    83: "Bi",
    84: "Po",
    85: "At",
    86: "Rn",
    87: "Fr",
    88: "Ra",
    89: "Ac",
    90: "Th",
    91: "Pa",
    92: "U",
    93: "Np",
    94: "Pu",
    95: "Am",
    96: "Cm",
    97: "Bk",
    98: "Cf",
    99: "Es",
    100: "Fm",
    101: "Md",
    102: "No",
    103: "Lr",
    104: "Rf",
    105: "Db",
    106: "Sg",
    107: "Bh",
    108: "Hs",
    109: "Mt",
    110: "Ds",
    111: "Rg",
    112: "Cn",
    113: "Nh",
    114: "Fl",
    115: "Mc",
    116: "Lv",
    117: "Ts",
    118: "Og",
}
SYMBOL_TO_NUM = {v: k for k, v in NUM_TO_SYMBOL.items()}


def _get_atomic_num(symbol):
    return SYMBOL_TO_NUM[symbol.lower().capitalize()]


def _parse_header(sdf):
    try:
        return sdf[0].strip(), sdf[1].strip(), sdf[2].strip()
    except IndexError:
        print(sdf)


def _parse_counts_line(sdf):
    # 12 fields
    # First 11 are 3 characters long
    # Last one is 6 characters long
    # First two give the number of atoms and bonds

    values = sdf[HEADER_SIZE]
    n_atoms = int(values[:3])
    n_bonds = int(values[3:6])

    return n_atoms, n_bonds


def _parse_atoms_block(sdf, n_atoms):
    # The first three fields, 10 characters long each, describe the atom's
    # position in the X, Y, and Z dimensions.
    # After that there is a space, and three characters for an atomic symbol.
    # After the symbol, there are two characters for the mass difference from
    # the monoisotope.
    # Next you have three characters for the charge.
    # There are ten more fields with three characters each, but these are all
    # rarely used.

    start = HEADER_SIZE + 1  # Add 1 for counts line
    stop = start + n_atoms
    values = sdf[start:stop]

    atoms = []
    for i, v in enumerate(values):
        coords = np.array([float(v[pos : pos + 10]) for pos in range(0, 30, 10)])
        atomic_num = _get_atomic_num(v[31:34].strip())
        iso = int(v[34:36])
        charge = int(v[36:39])
        info = np.array([int(v[pos : pos + 3]) for pos in range(39, len(v), 3)])
        atoms.append(
            {
                "index": i,
                "coords": coords,
                "atomic_num": atomic_num,
                "iso": iso,
                "charge": charge,
                "info": info,
            }
        )
    return atoms


def _parse_bonds_block(sdf, n_atoms, n_bonds):
    # The first two fields are the indexes of the atoms included in this bond
    # (starting from 1). The third field defines the type of bond, and the
    # fourth the stereoscopy of the bond.
    # There are a further three fields, with 3 characters each, but these are
    # rarely used and can be left blank.

    start = HEADER_SIZE + n_atoms + 1  # Add 1 for counts line
    stop = start + n_bonds
    values = sdf[start:stop]

    bonds = []
    for v in values:
        start_atom = int(v[:3]) - 1
        end_atom = int(v[3:6]) - 1
        type_ = int(v[6:9])
        stereo = int(v[9:12])
        info = np.array([int(v[pos : pos + 3]) for pos in range(12, len(v), 3)])
        bonds.append(
            {
                "start_atom": start_atom,
                "end_atom": end_atom,
                "type": type_,
                "stereo": stereo,
                "info": info,
            }
        )
    return bonds


def _parse_properties(sdf, n_atoms, n_bonds):
    # TODO This just returns a list of properties.
    # See https://docs.chemaxon.com/display/docs/MDL+MOLfiles%2C+RGfiles%2C+SDfiles%2C+Rxnfiles%2C+RDfiles+formats
    # for documentation.

    start = HEADER_SIZE + n_atoms + n_bonds + 1  # Add 1 for counts line
    stop = sdf.index("M  END")

    return sdf[start:stop]


def _parse_data_fields(sdf):
    # TODO This just returns a list of data fields.

    start = sdf.index("M  END") + 1

    return sdf[start:] if start < len(sdf) else []


def parse_sdf(sdf):
    sdf_out = {}
    sdf = sdf.split("\n")
    sdf_out["name"], sdf_out["details"], sdf_out["comment"] = _parse_header(sdf)
    sdf_out["n_atoms"], sdf_out["n_bonds"] = _parse_counts_line(sdf)
    sdf_out["atoms"] = _parse_atoms_block(sdf, sdf_out["n_atoms"])
    sdf_out["bonds"] = _parse_bonds_block(sdf, sdf_out["n_atoms"], sdf_out["n_bonds"])
    sdf_out["properties"] = _parse_properties(
        sdf, sdf_out["n_atoms"], sdf_out["n_bonds"]
    )
    sdf_out["data"] = _parse_data_fields(sdf)
    return sdf_out


def parse_sdf_file(sdf_file, amount=None):
    data = sdf_file.read().split("$$$$\n")
    if data[-1] == "":
        data = data[:-1]
    if amount is not None:
        data = data[:amount]
    output = [parse_sdf(sdf) for sdf in data]  # Parallel execution doesn't help
    return output


def load_sdf(filename, amount=None):
    """
    Load an .sdf file and return a list of molecules in the internal SDF format.
    :param filename: target SDF file
    :param amount: only load the first `amount` molecules from the file
    :return: a list of molecules in the internal SDF format (see documentation).
    """
    print("Reading SDF")
    with open(filename) as f:
        return parse_sdf_file(f, amount=amount)



from spektral.utils import normalized_adjacency


class NormalizeAdj:
    r"""
        Normalizes the adjacency matrix as:
    $$
        \A \leftarrow \D^{-1/2}\A\D^{-1/2}
        $$

        **Arguments**

        - `symmetric`: If False, then it computes \(\D^{-1}\A\) instead.
    """

    def __init__(self, symmetric=True):
        self.symmetric = symmetric

    def __call__(self, graph):
        if graph.a is not None:
            graph.a = normalized_adjacency(graph.a, self.symmetric)

        return graph



import tensorflow as tf
from scipy import sparse
from tensorflow.keras import backend as K

from spektral.layers import ops
from spektral.layers.pooling.src import SRCPool


class LaPool(SRCPool):
    r"""
    A Laplacian pooling (LaPool) layer from the paper

    > [Towards Interpretable Sparse Graph Representation Learning with Laplacian Pooling](https://arxiv.org/abs/1905.11577)<br>
    > Emmanuel Noutahi et al.

    **Mode**: disjoint.

    This layer computes a soft clustering of the graph by first identifying a set of
    leaders, and then assigning every remaining node to the cluster of the closest
    leader:
    $$
        \V = \|\L\X\|_d; \\
        \i = \{ i \mid \V_i > \V_j, \forall j \in \mathcal{N}(i) \} \\
        \S^\top = \textrm{SparseMax}\left( \beta \frac{\X\X_{\i}^\top}{\|\X\|\|\X_{\i}\|} \right)
    $$
    \(\beta\) is a regularization vecotr that is applied element-wise to the selection
    matrix.
    If `shortest_path_reg=True`, it is equal to the inverse of the shortest path between
    each node and its corresponding leader (this can be expensive since it runs on CPU).
    Otherwise it is equal to 1.

    The reduction and connection are computed as \(\X' = \S\X\) and
    \(\A' = \S^\top\A\S\), respectively.

    Note that the number of nodes in the output graph depends on the input node features.

    **Input**

    - Node features of shape `(n_nodes_in, n_node_features)`;
    - Adjacency matrix of shape `(n_nodes_in, n_nodes_in)`;

    **Output**

    - Reduced node features of shape `(n_nodes_out, channels)`;
    - Reduced adjacency matrix of shape `(n_nodes_out, n_nodes_out)`;
    - If `return_selection=True`, the selection matrix of shape
    `(n_nodes_in, n_nodes_out)`.

    **Arguments**

    - `shortest_path_reg`: boolean, apply the shortest path regularization described in
    the papaer (can be expensive);
    - `return_selection`: boolean, whether to return the selection matrix;
    """

    def __init__(self, shortest_path_reg=True, return_selection=False, **kwargs):
        super().__init__(return_selection=return_selection, **kwargs)

        self.shortest_path_reg = shortest_path_reg

    def call(self, inputs, **kwargs):
        x, a, i = self.get_inputs(inputs)

        # Select leaders
        lap = laplacian(a)
        v = ops.modal_dot(lap, x)
        v = tf.norm(v, axis=-1, keepdims=1)

        row = a.indices[:, 0]
        col = a.indices[:, 1]
        leader_check = tf.cast(tf.gather(v, row) >= tf.gather(v, col), tf.int32)
        leader_mask = ops.scatter_prod(leader_check[:, 0], row, self.n_nodes)
        leader_mask = tf.cast(leader_mask, tf.bool)

        return self.pool(x, a, i, leader_mask=leader_mask)

    def select(self, x, a, i, leader_mask=None):
        # Cosine similarity
        if i is None:
            i = tf.zeros(self.n_nodes, dtype=tf.int32)
        cosine_similarity = sparse_cosine_similarity(x, self.n_nodes, leader_mask, i)

        # Shortest path regularization
        if self.shortest_path_reg:

            def shortest_path(a_):
                return sparse.csgraph.shortest_path(a_, directed=False)

            np_fn_input = tf.sparse.to_dense(a) if K.is_sparse(a) else a
            beta = 1 / tf.numpy_function(shortest_path, [np_fn_input], tf.float64)
            beta = tf.where(tf.math.is_inf(beta), tf.zeros_like(beta), beta)
            beta = tf.boolean_mask(beta, leader_mask, axis=1)
            beta = tf.cast(
                tf.ensure_shape(beta, cosine_similarity.shape), cosine_similarity.dtype
            )
        else:
            beta = 1.0

        s = tf.sparse.softmax(cosine_similarity)
        s = beta * tf.sparse.to_dense(s)

        # Leaders end up entirely in their own cluster
        kronecker_delta = tf.boolean_mask(
            tf.eye(self.n_nodes, dtype=s.dtype), leader_mask, axis=1
        )

        # Create clustering
        s = tf.where(leader_mask[:, None], kronecker_delta, s)

        return s

    def reduce(self, x, s, **kwargs):
        return ops.modal_dot(s, x, transpose_a=True)

    def connect(self, a, s, **kwargs):
        return ops.matmul_at_b_a(s, a)

    def reduce_index(self, i, s, leader_mask=None):
        i_pool = tf.boolean_mask(i, leader_mask)

        return i_pool

    def get_config(self):
        config = {"shortest_path_reg": self.shortest_path_reg}
        base_config = super().get_config()
        return {**base_config, **config}


def laplacian(a):
    d = ops.degree_matrix(a, return_sparse_batch=True)
    if K.is_sparse(a):
        a = a.__mul__(-1)
    else:
        a = -a

    return tf.sparse.add(d, a)


def reduce_sum(x, **kwargs):
    if K.is_sparse(x):
        return tf.sparse.reduce_sum(x, **kwargs)
    else:
        return tf.reduce_sum(x, **kwargs)


def sparse_cosine_similarity(x, n_nodes, mask, i):
    mask = tf.cast(mask, tf.int32)
    leader_idx = tf.where(mask)

    # Number of nodes in each graph
    ns = tf.math.segment_sum(tf.ones_like(i), i)
    ks = tf.math.segment_sum(mask, i)

    # s will be a block-diagonal matrix where entry i,j is the cosine
    # similarity between node i and leader j.
    # The code below creates the indices of the sparse block-diagonal matrix
    # Row indices of the block-diagonal S
    starts = tf.cumsum(ns) - ns
    starts = tf.repeat(starts, ks)
    stops = tf.cumsum(ns)
    stops = tf.repeat(stops, ks)
    index_n = tf.ragged.range(starts, stops).flat_values

    # Column indices of the block-diagonal S
    index_k = tf.repeat(leader_idx, tf.repeat(ns, ks))
    index_k_for_s = tf.repeat(tf.range(tf.reduce_sum(ks)), tf.repeat(ns, ks))

    # Make index int64
    index_n = tf.cast(index_n, tf.int64)
    index_k = tf.cast(index_k, tf.int64)
    index_k_for_s = tf.cast(index_k_for_s, tf.int64)

    # Compute similarity between nodes and leaders
    x_n = tf.gather(x, index_n)
    x_n_norm = tf.norm(x_n, axis=-1)
    x_k = tf.gather(x, index_k)
    x_k_norm = tf.norm(x_k, axis=-1)
    values = tf.reduce_sum(x_n * x_k, -1) / (x_n_norm * x_k_norm)

    # Create a sparse tensor for S
    indices = tf.stack((index_n, index_k_for_s), 1)
    s = tf.SparseTensor(
        values=values, indices=indices, dense_shape=(n_nodes, tf.reduce_sum(ks))
    )
    s = tf.sparse.reorder(s)

    return s



from spektral import datasets
from spektral.data import BatchLoader, DisjointLoader, SingleLoader

batch_size = 3


def test_citation():
    dataset = datasets.Cora()
    dataset = datasets.Citeseer(random_split=True)
    dataset = datasets.Pubmed(normalize_x=True)
    sl = SingleLoader(dataset)
    sl.load()


def test_dblp():
    dataset = datasets.DBLP(normalize_x=True)
    sl = SingleLoader(dataset)
    sl.load()


def test_flickr():
    dataset = datasets.Flickr(normalize_x=True)
    sl = SingleLoader(dataset)
    sl.load()


def test_graphsage():
    # Test only PPI because Travis otherwise runs into memory errors
    dataset = datasets.PPI()
    _ = SingleLoader(dataset)


def test_mnist():
    _ = datasets.MNIST(k=8, noise_level=0.1)


def test_qm7():
    dataset = datasets.QM7()
    dl = DisjointLoader(dataset, batch_size=batch_size)
    dl.__next__()

    bl = BatchLoader(dataset, batch_size=batch_size)
    bl.__next__()


def test_qm9():
    dataset = datasets.QM9(amount=100)
    dl = DisjointLoader(dataset, batch_size=batch_size)
    dl.__next__()

    bl = BatchLoader(dataset, batch_size=batch_size)
    bl.__next__()


def test_tud():
    # Edge labels + edge attributes
    dataset = datasets.TUDataset("BZR_MD", clean=False)
    dl = DisjointLoader(dataset, batch_size=batch_size)
    dl.__next__()

    bl = BatchLoader(dataset, batch_size=batch_size)
    bl.__next__()

    # Node labels + node attributes + clean version
    dataset = datasets.TUDataset("ENZYMES", clean=True)
    dl = DisjointLoader(dataset, batch_size=batch_size)
    dl.__next__()

    bl = BatchLoader(dataset, batch_size=batch_size)
    bl.__next__()

    datasets.TUDataset.available_datasets
